/*****************************************************************************
 * Copyright (c) 2008, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "config.h"
#include "objs/kite_object.h"
#include "vm/kite_vm.h"

#ifndef WIN32
#include <unistd.h>
#else
#include <XGetopt.h>
#endif /* WIN32 */

#undef PACKAGE_NAME
#define PACKAGE_NAME "kdb"

/*****************************************************************************
 * Output software version
 ****************************************************************************/
void version() {
    printf("This is %s, version %s (built on %s)\n", PACKAGE_NAME, 
           PACKAGE_VERSION, TARGET_ARCH);
    printf("Copyright (C) 2009 Mooneer Salem. All rights reserved\n\n");

    /* BSD License text below copyright notice */
    printf("Redistribution and use in source and binary forms, with or without\n");
    printf("modification, are permitted provided that the following conditions are met:\n");
    printf("    * Redistributions of source code must retain the above copyright\n");
    printf("      notice, this list of conditions and the following disclaimer.\n");
    printf("    * Redistributions in binary form must reproduce the above copyright\n");
    printf("      notice, this list of conditions and the following disclaimer in the\n");
    printf("      documentation and/or other materials provided with the distribution.\n");
    printf("    * Neither the name of the Kite Language organization nor the\n");
    printf("      names of its contributors may be used to endorse or promote products\n");
    printf("      derived from this software without specific prior written permission.\n\n");
    printf("THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY\n");
    printf("EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED\n");
    printf("WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE\n");
    printf("DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY\n");
    printf("DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES\n");
    printf("(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;\n");
    printf("LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND\n");
    printf("ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT\n");
    printf("(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF\n");
    printf("THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n");
}

/*****************************************************************************
 * Output help message
 ****************************************************************************/
void help() {
    version();
    printf("\n");
    printf("Usage: %s [-h|-v] [file]\n", PACKAGE_NAME);
    printf("      -v: Print version information\n");
    printf("      -h: This message\n");
    printf("  [file]: File to execute or run (optional). If not specified,");
    printf(" defaults to standard input.\n\n");
}

/*****************************************************************************
 * Main application entry point.
 ****************************************************************************/
int main(int argc, char **argv) {
    int ch;
    char *file = NULL;
    kite_vm_t *vm;
    kite_thread_t *thd;
    kite_object_t *obj, *exc, *init_code;
    kite_funccall_info_t *fcall;
    char buf[2048];

#ifdef WIN32
    GetModuleFileName(NULL, buf, 2048);
    file = strstr(buf, "kdb.exe");
    file[0] = 0;
    strncat_s(file, 2048, "modules\\:.\\", 2048);
    kite_set_dylib_path(buf);
#endif /* WIN32 */

    kite_app_init();

    /* Parse command arguments first */
#ifdef __linux
    while ((ch = getopt(argc, argv, "+hv")) != -1) {
#else
    while ((ch = getopt(argc, argv, "hv")) != -1) {
#endif /* LINUX */
        switch(ch) {
            case 'v':
                version();
                exit(0);
                break;
            case 'h':
            default:
                help();
                exit(0);
                break;
        }
    }
    argc -= optind;
    argv += optind;
    
    file = argv[0];
    
    /* Load file into memory. */
    vm = kite_new_vm(argv);
    thd = kite_vm_creator_thread(vm);
    if (file != NULL) {
        obj = kite_vm_compile_from_file(thd, file);
    } else {
        obj = kite_vm_compile_from_fp(thd, stdin, "(stdin)", kite_new_class(
            thd, 
            kite_dereference_and_load(thd, "System.object"),
            ""
        ));
    }
    
    if (thd->exception) kite_vm_execute_exception(thd, TRUE);
    if (obj == NULL) {
#ifdef HAVE_GC_H
        kite_funccall_info_t *entry = (void*)GC_malloc(sizeof(kite_funccall_info_t));
#else
        kite_funccall_info_t *entry = calloc(sizeof(kite_funccall_info_t), 1);
#endif /* HAVE_GC_H */
        entry->this = NULL;
        entry->file = "(Kite loader)";
        kite_push_stack(&thd->func_stack, FALSE, entry);
                
        exc = 
            kite_new_exception(thd, "System.exceptions.PackageNotFound", 
                               file);
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        kite_vm_execute_exception(thd, TRUE);
    }
    
    /* Load debugger module (System.vm.debugger) and instrument. */
#ifndef HAVE_GC_H
    fcall = (kite_funccall_info_t*)malloc(sizeof(kite_funccall_info_t));
#else
    fcall = (kite_funccall_info_t*)GC_malloc(sizeof(kite_funccall_info_t));
#endif /* HAVE_GC_H */
    fcall->file = "(kdb instrumentation)";
    fcall->line = 0;
    fcall->locals = NULL;
    fcall->sys_or_user = 0;
    fcall->this = obj;
    kite_push_stack(&thd->func_stack, FALSE, fcall);
    
    init_code = kite_vm_compile_from_string(thd,
        "import \"System.vm.debugger\"; _dbg = make System.vm.debugger(); __root|__instruction_tracer(_dbg.stepper);", 1);
    kite_vm_execute_user_method(thd, init_code->builtin_data.funcvalue.funcptr);

    if (thd->exception) kite_vm_execute_exception(thd, TRUE);
    
    kite_pop_stack(&thd->func_stack);
    free(fcall);

    kite_start_bytecode(thd, obj);
    
    /* Wait for thread to stop, then free everything and quit. */
    kite_join_thread(vm, thd);
    kite_free_vm(&vm);
    
    return 0;
}
