#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "kite_object.h"
#include "kite_vm.h"

#define OBJ_ARRAY OBJ_INSTANCE + 1

#define PERFORM_ARRAY_BOOLEAN_OP(op, c) \
    kite_object_t *obj,*rhs,*exc; \
    int idx, max; \
    if (args->type == OBJ_LIST) { \
        rhs = args->builtin_data.listvalue.car; \
    } else { \
        rhs = args; \
    } \
    if (kite_array_size(this) != kite_array_size(rhs)) { \
      if (kite_array_size(this) op kite_array_size(rhs)) { \
        kite_vm_return(thd, kite_new_boolean(thd, TRUE)); } \
      else { \
        kite_vm_return(thd, kite_new_boolean(thd, FALSE)); } \
      return; \
    } \
    max = kite_array_size(this); \
    for (idx=0; idx<max; idx++) { \
        kite_vm_call_operator(thd, kite_array_get(this,idx), \
            c, kite_array_get(rhs,idx), TRUE); \
        obj = kite_vm_pop(thd); \
        \
        if (obj->type == OBJ_BOOLEAN) { \
            if (obj->builtin_data.intvalue == 0) { \
                kite_dereference_object(obj); \
                kite_vm_return(thd, kite_new_boolean(thd, FALSE)); \
            } \
            kite_dereference_object(obj); \
        } else { \
            exc = kite_new_exception(thd, "System.exceptions.InvalidArgument", \
                                "Comparison did not result in a boolean."); \
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
            return; \
        } \
        \
    } \
    \
    kite_vm_return(thd, kite_new_boolean(thd, TRUE));

typedef struct {
  long size /* size of the array */;
  long capacity /* capacity */;
  kite_object_t **array /* the array itself */;
} array_data;

void kite_array_reallocate(kite_object_t *this, long space);
void kite_array_add(kite_object_t *this, kite_object_t *item, long at);
void kite_array_append_item(kite_object_t *this, kite_object_t *item);
void kite_array_append_array(kite_object_t *this, kite_object_t *other);
kite_object_t* kite_array_get(kite_object_t *this, long at);
array_data *kite_new_array_struct(int size);
kite_object_t *kite_new_array(kite_thread_t *thd, int size);
kite_object_t *kite_sort_array(kite_thread_t *thd, kite_object_t *array_object, kite_object_t *method);

int kite_array_size(kite_object_t *array) {
  return ((array_data *)(array->builtin_data.filevalue))->size;
}

KITE_CLASS_METHOD(Array_lessthan) {
    PERFORM_ARRAY_BOOLEAN_OP(<, OP_LESS_THAN);
}

KITE_CLASS_METHOD(Array_lessequals) {
    PERFORM_ARRAY_BOOLEAN_OP(<=, OP_LESS_OR_EQUALS);
}

KITE_CLASS_METHOD(Array_equals) {
    PERFORM_ARRAY_BOOLEAN_OP(==, OP_EQUALS);
}

KITE_CLASS_METHOD(Array_greaterequals) {
    PERFORM_ARRAY_BOOLEAN_OP(>=, OP_GREATER_OR_EQUALS);
}

KITE_CLASS_METHOD(Array_greaterthan) {
    PERFORM_ARRAY_BOOLEAN_OP(>, OP_GREATER_THAN);
}

KITE_CLASS_METHOD(Array_notequals) {
    PERFORM_ARRAY_BOOLEAN_OP(!=, OP_NOT_EQUALS);
}

/* Allocate space for the struct, then 
 * assign it to the builtin_data's filevalue.
 */
KITE_CLASS_METHOD(Array_construct) {
  array_data *array_struct = kite_new_array_struct(0);
  
  KITE_NO_ARGS;
  (void)thd;
  
  this->builtin_data.filevalue = (FILE *)array_struct;
  this->type = OBJ_ARRAY;
}

KITE_CLASS_METHOD(Array_capacity) {
  array_data *mystruct = (array_data *)(this->builtin_data.filevalue);
  KITE_NO_ARGS;
  kite_vm_return(thd, kite_new_integer(thd, mystruct->capacity));
}

KITE_CLASS_METHOD(Array_size) {
  array_data *mystruct = (array_data *)(this->builtin_data.filevalue);
  KITE_NO_ARGS;
  kite_vm_return(thd, kite_new_integer(thd, mystruct->size));
}


// kite_array_add
KITE_CLASS_METHOD(Array_set) {
  kite_object_t *item;
  kite_object_t *location;
  
  KITE_GET_METHOD_ARGUMENT(location, 1); // now location has the second argument
  KITE_GET_METHOD_ARGUMENT(item, 2); // now item has the first argument
  
  
  // and the actual magic
  kite_array_add(this, item, location->builtin_data.intvalue);
  
  kite_vm_return(thd, this); // return the ARRAY not the item passed
}

// kite_array_append_item
KITE_CLASS_METHOD(Array_get_item) {
  kite_object_t *index,*result;
  long at;
  array_data *myarray = (array_data *)this->builtin_data.filevalue;
  
  if (args->type == OBJ_LIST) { 
      index = args->builtin_data.listvalue.car;
  } else {
      index = args;
  }
  at = KITE_GET_INTEGER(index);
  
  if (at < 0 || at >= myarray->capacity) {
    kite_vm_return(thd,kite_new_null(thd));
    return;
  }
  result = kite_array_get(this,at);
  if (result == NULL)
  {
    kite_vm_return(thd,kite_new_null(thd));
  }
  else
  {
    kite_vm_push(thd,result);
    kite_vm_return(thd, result);
  } 
}

KITE_CLASS_METHOD(Array_sort) {
  kite_object_t *method,*result;
  
  if (args->type == OBJ_LIST) { 
      method = args->builtin_data.listvalue.car;
  } else {
      method = args;
  }
  result = kite_sort_array(thd,this,method);
  kite_vm_return(thd,kite_reference_object(result));
}

KITE_CLASS_METHOD(Array_reverse) {
  int i;
  kite_object_t *newarr;
  array_data *array_struct = (array_data *)this->builtin_data.filevalue;

  KITE_NO_ARGS;  
  newarr = kite_new_array(thd,0);
  
  for (i=array_struct->size-1;i>=0;i--) {
    kite_object_t *copied_object = kite_array_get(this,i);
    kite_array_append_item(newarr,kite_reference_object(copied_object));
  }
  kite_vm_return(thd,kite_reference_object(newarr));
}

KITE_CLASS_METHOD(Array_print) {
  int idx = 0;
  array_data *array_struct = (array_data *)this->builtin_data.filevalue;
  kite_object_t *stringrep;
  char *ret = strdup("[");
  
  KITE_NO_ARGS;

  for(idx=0;array_struct && idx < array_struct->size;idx++) {
      stringrep = kite_string_object(thd, array_struct->array[idx]);
      ret = realloc(ret, strlen(ret) + 
          strlen(stringrep->builtin_data.stringvalue.string) + 3);
      strcat(ret, stringrep->builtin_data.stringvalue.string);
      kite_dereference_object(stringrep);
      if (idx < array_struct->size-1) {
          strcat(ret, ", ");
      }
  }
  
  strcat(ret, "]");
  printf("%s\n",ret);
  free(ret);
  kite_vm_return(thd, this);
  
}

KITE_CLASS_METHOD(Array_str) {
  int idx = 0;
  array_data *array_struct = (array_data *)this->builtin_data.filevalue;
  kite_object_t *stringrep;
  char *ret = strdup("[");
  
  KITE_NO_ARGS;

  for(idx=0;idx < array_struct->size;idx++) {
      stringrep = kite_string_object(thd, array_struct->array[idx]);
      ret = realloc(ret, strlen(ret) + 
          strlen(stringrep->builtin_data.stringvalue.string) + 3);
      strcat(ret, stringrep->builtin_data.stringvalue.string);
      kite_dereference_object(stringrep);
      if (idx < array_struct->size-1) {
          strcat(ret, ", ");
      }
  }
  
  strcat(ret, "]");
  stringrep = kite_new_string(thd, ret);
  free(ret);
  kite_vm_return(thd, stringrep);
}


KITE_CLASS_METHOD(Array_reduce) {
  int idx = 0;  
  kite_object_t *method,*result;
  array_data *array_struct = (array_data *)this->builtin_data.filevalue;
  
  if (args->type == OBJ_LIST) { 
      method = args->builtin_data.listvalue.car;
  } else {
      method = args;
  }
  result = kite_new_null(thd);

  for (idx=0;idx<array_struct->size;idx++)
  {
    kite_object_t *cur = kite_array_get(this,idx);
    kite_object_t *args = kite_new_list(thd);
    kite_append_list(thd,args,result);
    kite_append_list(thd,args,cur);
    
    kite_vm_call_object(thd, kite_new_null(thd), method, args);
    
    kite_dereference_object(args);
    result = kite_vm_pop(thd);
    
  }
  
  kite_vm_return(thd,kite_reference_object(result));
}
  
KITE_CLASS_METHOD(Array_map) {
  int idx = 0;  
  kite_object_t *newarr,*method;
  array_data *array_struct = (array_data *)this->builtin_data.filevalue;
  
  newarr = kite_new_array(thd,0);
  
  if (args->type == OBJ_LIST) { 
      method = args->builtin_data.listvalue.car;
  } else {
      method = args;
  }

  for (idx=0;idx<array_struct->size;idx++)
  {
    kite_object_t *cur = kite_array_get(this,idx);
    kite_object_t *args = kite_list_object(thd,cur);
    
    kite_vm_call_object(thd, kite_new_null(thd), method, args);
    
    kite_dereference_object(args);
    kite_array_append_item(newarr, kite_reference_object(kite_vm_pop(thd)));
  }
  
  kite_vm_return(thd,kite_reference_object(newarr));
}

KITE_CLASS_METHOD(Array_index_of) {
  int idx = 0;
  kite_object_t *lookup;
  array_data *array_struct = (array_data *)this->builtin_data.filevalue;
  
  lookup = args->builtin_data.listvalue.car;
  
  for (idx=0;idx<array_struct->size;idx++)
  {
    kite_object_t *obj,*cur = kite_array_get(this,idx);
    kite_vm_call_operator(thd, cur, OP_EQUALS, lookup, TRUE);
    obj = kite_vm_pop(thd);
    
    if (obj->type == OBJ_BOOLEAN) {
      if (obj->builtin_data.intvalue == 1) {
        kite_vm_return(thd,kite_new_integer(thd,idx));
        return;
      }
    }
    kite_dereference_object(obj);
    
  }
  kite_vm_return(thd,kite_new_integer(thd,-1));
}

// kite_array_append_item
KITE_CLASS_METHOD(Array_append_item) {
  kite_object_t *rhs;
  if (args->type == OBJ_LIST) { 
      rhs = args->builtin_data.listvalue.car;
  } else {
      rhs = args;
  }
  
  // and the actual magic
  kite_array_append_item(this, rhs);
  
  kite_vm_return(thd, this); // return the ARRAY not the item passed
}


// kite_array_append_array
KITE_CLASS_METHOD(Array_append_array) {
  kite_object_t *array;
  
  KITE_GET_METHOD_ARGUMENT(array, 1); // now item has the first argument
  
  // and the actual magic
  if (array->type == OBJ_ARRAY)
    kite_array_append_array(this, array);
  else
    kite_array_append_item(this,array);
  
  kite_vm_return(thd, this); // return the ARRAY not the item passed
}

/*****************************************
 * IMPLEMENTATIONS OF THE ABOVE
 */


/* Reallocates another +space+ slots
 * 
 * @param this self, which will have its array reallocated
 */
 
kite_object_t *kite_sort_array(kite_thread_t *thd, kite_object_t *array_object, kite_object_t *method) {
  int start = 0, end = ((array_data *)(array_object->builtin_data.filevalue))->size;
  int idx;
  kite_object_t **array, *partition, *left, *right, *result;
  
  if (end - start <= 1) return array_object;
  array = ((array_data *)(array_object->builtin_data.filevalue))->array;
  partition = array[start];
  left = kite_new_array(thd,0); right = kite_new_array(thd,0);

  for (idx=start+1;idx<end;idx++) {
    kite_object_t *cur = array[idx];
    kite_object_t *args = kite_new_list(thd);
    long result;
    
    kite_append_list(thd,args,partition);
    kite_append_list(thd,args,array[idx]);
    
    kite_vm_call_object(thd, kite_new_null(thd), method, args);
    result = KITE_GET_INTEGER(kite_vm_pop(thd));
    
    switch (result) {
      case -1:
        kite_array_append_item(left,cur);
        break;
      case 0:
        kite_array_append_item(right,cur);
        break;
      case 1:
        kite_array_append_item(right,cur);
        break;
    }
  }
  result = kite_new_array(thd,0);
  
  left = kite_sort_array(thd,left,method);
  right = kite_sort_array(thd,right,method);
  
  kite_array_append_array(result,left);
  kite_array_append_item(result,partition);
  kite_array_append_array(result,right);
  
  return result;
  
}
 
kite_object_t *kite_new_array(kite_thread_t *thd, int size) {
  array_data *new_array_struct;
  kite_object_t *newarr;
  kite_object_t *arrayclass = 
    kite_dereference_and_load(thd, "array");
  newarr = kite_new_instance(thd,arrayclass);
  
  new_array_struct = kite_new_array_struct(size);
  newarr->builtin_data.filevalue = (FILE *)new_array_struct;
  newarr->type = OBJ_ARRAY;
  return kite_reference_object(newarr);
} 

array_data *kite_new_array_struct(int size) {
  array_data *array_struct = calloc(1, sizeof(array_data)); /* 100 is the starting size */
  int newcapacity = ((size / 100) + 1) * 100;
  
  array_struct->array = calloc(newcapacity,sizeof(kite_object_t*));
  array_struct->capacity = newcapacity;
  array_struct->size = 0;
  return array_struct;
}

void kite_array_reallocate(kite_object_t *this, long space) {
  array_data *array_struct = (array_data *)(this->builtin_data.filevalue);
  
  // add `space` slots to it
  array_struct->array = realloc(array_struct->array, (array_struct->capacity + space) * sizeof(kite_object_t *));
  array_struct->capacity = (array_struct->capacity + space);
}


/* Add an element +item+ to the array at slot +at+.
 * If +at+ is outside the boundaries of the array, it will dynamically
 * reallocate another 100 slots.
 * 
 * @param this self
 * @param item the object you're adding
 * @param at   where we're adding the item
 */
void kite_array_add(kite_object_t *this, kite_object_t *item, long at) {
  array_data *array_struct = (array_data *)this->builtin_data.filevalue; // the struct which has the array
  
  // reallocate if they're accessing something outside of the array boundaries
  // this allows you to do array = []; array|set_value(140, "hi")
  long newcapacity = ((at / 100) + 1) * 100; //round to the nearest 100th
  
  if (at >= array_struct->capacity) {
    kite_array_reallocate(this,newcapacity - array_struct->capacity);
  }
  kite_reference_object(item);
  array_struct->array[at] = item; // add the item
  if (at+1 > array_struct->size)
    array_struct->size = at+1; // update the size
}


/* Append an array +other+ onto this one.
 * 
 * @param this  self
 * @param other the array we're adding
 */
void kite_array_append_array(kite_object_t *this, kite_object_t *other) {
  long next = 100; /* round up to the next multiple of this */
  
  array_data *array_struct = (array_data *)this->builtin_data.filevalue; // the struct which has the array
  array_data *other_struct = (array_data *)other->builtin_data.filevalue;
  
  /* new capacity is the next hundred higher */
  long new_capacity = (long)((((array_struct->size + other_struct->size) / next) +1) * next); // the next multiple of +next+
  long difference   = (long)(new_capacity - array_struct->capacity);
  
  int start = 0, end, idx;
  
  /* add the sizes
   * round up to the next 100
   */
  kite_array_reallocate(this, difference);
  end = other_struct->size;
  for (idx=0;idx<end;idx++) {
    kite_array_append_item(this,other_struct->array[idx]);
  }
}


/* Put +item+ in the next available slot.
 * 
 * @param this  self
 * @param other the item we're appending
 */
void kite_array_append_item(kite_object_t *this, kite_object_t *item) {
  array_data *array_struct = (array_data *)this->builtin_data.filevalue; // the struct which has the array
  
  kite_array_add(this, item, array_struct->size); // add it to the next available slot, which is always `size`
}

kite_object_t* kite_array_get(kite_object_t *this, long at) {
  array_data *array_struct = (array_data *)this->builtin_data.filevalue; // the struct which has the array
  
  return array_struct->array[at]; // add it to the next available slot, which is always `size`
}

// moving them into the class
KITE_MODULE_INITIALIZER(array) {
  /*
   * Create new class object. inherits is typically NULL for packages
   * and System.object.
   */
  kite_object_t *inherits = 
    kite_dereference_and_load(thread, "System.object");
  kite_object_t *newclass = 
    kite_new_class(thread, inherits, "array");
  kite_set_docstring(newclass, "Object representing an array of objects.");
  
  kite_add_operator(thread, newclass, OP_REDUCE, 
    kite_new_method_compiled_with_docs(thread, Array_reduce, 
    "Reduces an array by calling the method provided on every item, and passing the result to the next item. E.g.: ['Hello',', World!'] <| method(result,item) [ result + item; ] = 'Hello, World'", 1,
    "method", "The method to run on every item in the array."));  
    
  kite_add_operator(thread, newclass, OP_MAP, 
    kite_new_method_compiled_with_docs(thread, Array_map, 
                                "Runs the given method to every item in the array, and returns a new array with the resulting objects", 1,
                                "method", "The method to run on every item in the array."));  
    
   kite_add_operator(thread, newclass, OP_LEFT_SHIFT, 
     kite_new_method_compiled_with_docs(thread, Array_append_item, 
                               "Appends item to the end of this array", 1,
                               "item", "item to append."));
     
  kite_add_operator(thread, newclass, OP_ADD, 
    kite_new_method_compiled_with_docs(thread, Array_append_array, 
                                "Appends array to the end of this array", 1,
                                "array", "Array to append."));

  kite_add_operator(thread, newclass, OP_ARRAY_DEREF, 
    kite_new_method_compiled_with_docs(thread, Array_get_item, 
                                "Gets the item at the given index", 1,
                                "index", "Index to lookup."));
  kite_add_operator(thread, newclass, OP_LESS_THAN, 
      kite_new_method_compiled_with_docs(thread, Array_lessthan, 
                                "Returns true if every item in the left array is less than every item in the right.", 1,
                                "rhs", "The array to compare to."));
  kite_add_operator(thread, newclass, OP_LESS_OR_EQUALS, 
      kite_new_method_compiled_with_docs(thread, Array_lessequals,
                                "Returns true if every item in the left array is less than or eaual to every item in the right.", 1,
                                "rhs", "The array to compare to."));
  kite_add_operator(thread, newclass, OP_EQUALS, 
      kite_new_method_compiled_with_docs(thread, Array_equals, 
                                "Compares two arrays for equality.", 1,
                                "rhs", "The array to compare to."));
  kite_add_operator(thread, newclass, OP_GREATER_OR_EQUALS, 
      kite_new_method_compiled_with_docs(thread, Array_greaterequals,
                                "Returns true if every item in the left array is greater than or equal to every item in the right.", 1,
                                "rhs", "The array to compare to."));
  kite_add_operator(thread, newclass, OP_GREATER_THAN, 
      kite_new_method_compiled_with_docs(thread, Array_greaterthan,
                                "Returns true if every item in the left array is greater than every item in the right.", 1,
                                "rhs", "The array to compare to."));
  kite_add_operator(thread, newclass, OP_NOT_EQUALS, 
      kite_new_method_compiled_with_docs(thread, Array_notequals, 
                                "Compares two arrays for lack of equality.", 1,
                                "rhs", "The array to compare to."));
  
  
  kite_add_method(thread, newclass, "sort",
      kite_new_method_compiled_with_docs(thread, Array_sort,
                                "Sorts the array.", 1,
                                "comparator","The method used to sort the objects."));
                                    
  kite_add_method(thread, newclass, "reverse",
      kite_new_method_compiled_with_docs(thread, Array_reverse, 
                                "Reverses the array. Object returned is a new array.", 0)); 
  kite_add_method(thread, newclass, "map",
      kite_new_method_compiled_with_docs(thread, Array_map,
                                "Maps the method onto the array", 1,
                                "method", "The method to apply to every item in the array"));
  
  kite_add_method(thread, newclass, "capacity",
      kite_new_method_compiled_with_docs(thread, Array_capacity, 
                                "Returns the current capacity of the array. Only interesting internally.", 0));
                                
  kite_add_method(thread, newclass, "append_item",
      kite_new_method_compiled_with_docs(thread, Array_append_item, 
                                "Appends an item to the end of the array.", 1,
                                "item","the item to put at the end of the array"));
                                
  kite_add_method(thread, newclass, "print",
      kite_new_method_compiled_with_docs(thread, Array_print, 
                                "Prints the array.", 0));
                                
  kite_add_method(thread, newclass, "str",
      kite_new_method_compiled_with_docs(thread, Array_str, 
                                "Returns the string form of the array.", 0));
                                
  kite_add_method(thread, newclass, "index_of",
      kite_new_method_compiled_with_docs(thread, Array_index_of, 
                                "Finds the index of a given object in the array, or -1 if not found.", 1,
                                "object", "Object to lookup."));
                                
  kite_add_method(thread, newclass, "get",
      kite_new_method_compiled_with_docs(thread, Array_get_item, 
                                "Gets the item at the given index", 1,
                                "index", "Index to lookup."));

  kite_add_method(thread, newclass, "set",
      kite_new_method_compiled_with_docs(thread, Array_set, 
                                "Sets the given index to the provided value.", 2,
                                "index","the index to use",
                                "item","item to add"));  
                                
  kite_add_method(thread, newclass, "size",
      kite_new_method_compiled_with_docs(thread, Array_size, 
                                "Returns the number of objects in the array.", 0));
                                  
  kite_add_method(thread, newclass, "__construct__",
      kite_new_method_compiled_with_docs(thread, Array_construct, 
                                "Creates an empty array.", 0));
  /*
   * Add operators
   */
   

    
  /* Add to parent to complete loading. */
  kite_add_property(thread, parent, "array", 1, NULL);
  kite_set_property(thread, parent, "array", newclass);


}
