//gcc -fPIC -dynamiclib -Wl,-undefined,dynamic_lookup -o hires.dylib hires.c
#include <stdlib.h>
#include <string.h>
#ifndef WIN32
#include <unistd.h>
#include <sys/time.h>
#else
#include <time.h>
#include <Winsock2.h>
#endif /* WIN32 */
#include <errno.h>
#define __USE_XOPEN

#ifdef WIN32
#if defined(_MSC_VER) || defined(_MSC_EXTENSIONS)
  #define DELTA_EPOCH_IN_MICROSECS  11644473600000000Ui64
#else
  #define DELTA_EPOCH_IN_MICROSECS  11644473600000000ULL
#endif
 
struct timezone 
{
  int  tz_minuteswest; /* minutes W of Greenwich */
  int  tz_dsttime;     /* type of dst correction */
};
 
int gettimeofday(struct timeval *tv, struct timezone *tz)
{
  FILETIME ft;
  unsigned __int64 tmpres = 0;
  static int tzflag;
 
  if (NULL != tv)
  {
    GetSystemTimeAsFileTime(&ft);
 
    tmpres |= ft.dwHighDateTime;
    tmpres <<= 32;
    tmpres |= ft.dwLowDateTime;
 
    /*converting file time to unix epoch*/
    tmpres /= 10;  /*convert into microseconds*/
    tmpres -= DELTA_EPOCH_IN_MICROSECS; 
    tv->tv_sec = (long)(tmpres / 1000000UL);
    tv->tv_usec = (long)(tmpres % 1000000UL);
  }
 
  if (NULL != tz)
  {
    if (!tzflag)
    {
      _tzset();
      tzflag++;
    }
    tz->tz_minuteswest = _timezone / 60;
    tz->tz_dsttime = _daylight;
  }
 
  return 0;
}
#endif /* WIN32 */

#include "kite_object.h"
#include "kite_vm.h"

#define GET_BREAKDOWN_FIELD(field) \
    struct tm *thistime; \
    time_t curtime; \
    KITE_NO_ARGS; \
    curtime = KITE_GET_INTEGER(kite_get_property(this,"_sec")); \
    thistime = gmtime(&curtime); \
    kite_vm_return(thd, kite_new_integer(thd, thistime->field));

#define TYPECHECK_OPERANDS \
    kite_object_t *rhs; \
    kite_object_t *exc; \
    if (args->type == OBJ_LIST) { \
        rhs = args->builtin_data.listvalue.car; \
    } else { \
        rhs = args; \
    } \
    if (this->type != rhs->type) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Both operands must be of the same type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }


/* Subtract the `struct timeval' values X and Y,
   storing the result in RESULT.
   Return 1 if the difference is negative, otherwise 0.  */

int
timeval_subtract (result, x, y)
     struct timeval *result, *x, *y;
{
  /* Perform the carry for the later subtraction by updating y. */
  if (x->tv_usec < y->tv_usec) {
    int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
    y->tv_usec -= 1000000 * nsec;
    y->tv_sec += nsec;
  }
  if (x->tv_usec - y->tv_usec > 1000000) {
    int nsec = (y->tv_usec - x->tv_usec) / 1000000;
    y->tv_usec += 1000000 * nsec;
    y->tv_sec -= nsec;
  }

  /* Compute the time remaining to wait.
     tv_usec is certainly positive. */
  result->tv_sec = x->tv_sec - y->tv_sec;
  result->tv_usec = x->tv_usec - y->tv_usec;

  /* Return 1 if result is negative. */
  return x->tv_sec < y->tv_sec;
}


static void 
Date_copyToProperties(kite_thread_t *thd, kite_object_t *obj, struct tm *t)
{
    time_t result = mktime(t);
    kite_set_property(thd, obj, "_sec", kite_new_integer(thd, result));
    kite_set_property(thd, obj, "_usec", kite_new_integer(thd, 0));
}

KITE_CLASS_METHOD(hires_subtract) {
  struct timeval *me = malloc(sizeof(struct timeval));
  struct timeval *you = malloc(sizeof(struct timeval));
  struct timeval *result = malloc(sizeof(struct timeval));
  
  TYPECHECK_OPERANDS;
    
  me->tv_sec = KITE_GET_INTEGER(kite_get_property(this,"_sec"));
  me->tv_usec = KITE_GET_INTEGER(kite_get_property(this,"_usec"));
    
     you->tv_sec = KITE_GET_INTEGER(kite_get_property(rhs,"_sec"));
    you->tv_usec = KITE_GET_INTEGER(kite_get_property(rhs,"_usec"));
    
  timeval_subtract(result,me,you);
  
  kite_vm_return(thd, kite_new_float(thd, result->tv_sec + (result->tv_usec * (1.0f/1000000))));
  // kite_vm_return(thd,kite_new_null(thd));
}

KITE_CLASS_METHOD(hires_nowLocal)
{
  kite_object_t *newObject;
  struct timeval *now;
  
  KITE_NO_ARGS;
  KITE_THIS_NOT_USED;
    
  newObject = kite_dereference_and_load(thd, "hires");
  newObject = kite_new_instance_with_constructor(
      thd, newObject, kite_new_list(thd));
      
  //grab the file argument
  now = malloc(sizeof(struct timeval));
  
  gettimeofday(now,NULL);
  kite_set_property(thd,newObject,"_sec",kite_new_integer(thd,now->tv_sec));
  kite_set_property(thd,newObject,"_usec",kite_new_integer(thd,now->tv_usec));
  
  kite_vm_push(thd, newObject);
  kite_dereference_object(newObject);
  free(now);
}

KITE_CLASS_METHOD(hires_nowGmt)
{
  kite_object_t *newObject;
  struct timeval *now;
  struct timezone *zone;
  
  KITE_NO_ARGS;
  KITE_THIS_NOT_USED;
  
  newObject = kite_dereference_and_load(thd, "hires");
  newObject = kite_new_instance_with_constructor(
      thd, newObject, kite_new_list(thd));
      
  //grab the file argument
  now = malloc(sizeof(struct timeval));
  zone = malloc(sizeof(struct timezone));
  gettimeofday(now,zone);
  
  kite_set_property(thd,newObject,"_sec",kite_new_integer(thd,now->tv_sec + (zone->tz_minuteswest * 60)));
  kite_set_property(thd,newObject,"_usec",kite_new_integer(thd,now->tv_usec));
  
  kite_vm_push(thd, newObject);
  kite_dereference_object(newObject);
  free(now);
  free(zone);
}

KITE_CLASS_METHOD(hires_str)
{
    time_t t = (time_t)KITE_GET_INTEGER(kite_get_property(this,"_sec"));
    kite_object_t *ret;

    KITE_NO_ARGS;
        
    ret = kite_new_string(thd, ctime(&t));
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}
KITE_CLASS_METHOD(hires_usec)
{
  time_t usec = (time_t)KITE_GET_INTEGER(kite_get_property(this,"_usec")); 
  KITE_NO_ARGS;
  kite_vm_return(thd, kite_new_integer(thd, usec));
}

KITE_CLASS_METHOD(hires_sec)
{
  GET_BREAKDOWN_FIELD(tm_sec);
}

KITE_CLASS_METHOD(hires_min)
{
  GET_BREAKDOWN_FIELD(tm_min);
}

KITE_CLASS_METHOD(hires_hour)
{
  GET_BREAKDOWN_FIELD(tm_hour);
}

KITE_CLASS_METHOD(hires_day_of_month)
{
  GET_BREAKDOWN_FIELD(tm_mday);
}

KITE_CLASS_METHOD(hires_month)
{
  GET_BREAKDOWN_FIELD(tm_mon);
}

KITE_CLASS_METHOD(hires_year)
{
  GET_BREAKDOWN_FIELD(tm_year);
}

KITE_CLASS_METHOD(hires_timestamp)
{
    kite_object_t *ret;

    KITE_NO_ARGS;

    ret = kite_new_integer(thd, KITE_GET_INTEGER(kite_get_property(this,"_sec")));
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}

KITE_CLASS_METHOD(hires_construct)
{
  KITE_NO_ARGS;
  kite_set_property(thd, this, "_sec", kite_new_integer(thd, 0));
  kite_set_property(thd, this, "_usec", kite_new_integer(thd, 0));
}
/*****************************************************************************
 * Format date using given format string
 ****************************************************************************/
KITE_CLASS_METHOD(hires_format)
{
    struct tm *thistime;
    time_t curtime;
    kite_object_t *arg, *ret;
    char buf[2048];
    
    KITE_GET_METHOD_ARGUMENT(arg, 1);
    if (!KITE_IS_TYPE(arg, OBJ_STRING)) {
        kite_object_t *exc;
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch",
                                "Operand must be string.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }
    curtime = KITE_GET_INTEGER(kite_get_property(this,"_sec"));
    thistime = gmtime(&curtime);
    strftime(buf, 2048, KITE_GET_STRING_VALUE(arg), thistime);
    
    ret = kite_new_string(thd, buf);
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}
/*****************************************************************************
 * Parse date using given format string
 ****************************************************************************/
KITE_CLASS_METHOD(hires_parse)
{
    kite_object_t *date, *format;
    kite_object_t *newObject;
    struct tm *current = malloc(sizeof(struct tm));

    (void)this;
    KITE_GET_METHOD_ARGUMENT(date, 1);
    KITE_GET_METHOD_ARGUMENT(format, 2);
    
    if (!KITE_IS_TYPE(date, OBJ_STRING) ||
        !KITE_IS_TYPE(format, OBJ_STRING)) {
        kite_object_t *exc;
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch",
                                "Operand must be string.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }
    
    
    newObject = kite_dereference_and_load(thd, "hires");
    newObject = kite_new_instance_with_constructor(
        thd, newObject, kite_new_list(thd));
    
    strptime(KITE_GET_STRING_VALUE(date), 
             KITE_GET_STRING_VALUE(format), current);
    Date_copyToProperties(thd, newObject, current);
    kite_vm_push(thd, newObject);
    kite_dereference_object(newObject);
    free(current);
}
KITE_MODULE_INITIALIZER(hires)
{
   kite_object_t *inherits =
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass =
       kite_new_class(thread, inherits, "hires");
   kite_set_docstring(newclass, "Class for representing a time of day.");
   
   kite_add_operator(thread, newclass, OP_SUBTRACT, 
       kite_new_method_compiled_with_docs(thread, hires_subtract, 
                                 "Subtracts two dates with fractional-second precision.", 1,
                                 "rhs", "The date to subtract from the current date."));
   
   kite_add_method(thread, newclass, "nowLocal",
       kite_new_method_compiled_with_docs(thread, hires_nowLocal,
                                 "Returns the current local time.", 0));

   kite_add_method(thread, newclass, "nowGmt",
       kite_new_method_compiled_with_docs(thread, hires_nowGmt,
                                 "Returns the current time, adjusted for GMT.", 0));
   
   kite_add_method(thread, newclass, "str",
       kite_new_method_compiled_with_docs(thread, hires_str,
                                 "Provides the time in a readable format.", 0));
                                 
   kite_add_method(thread, newclass, "timestamp",
       kite_new_method_compiled_with_docs(thread, hires_timestamp,
                                 "Return given date as the number of seconds since 1970.", 0));
   kite_add_method(thread, newclass, "parse", 
       kite_new_method_compiled_with_docs(thread, hires_parse, 
                                 "Parse given value into a date object.", 2,
                                 "val", "Value to parse.",
                                 "format", "Format to use (see man strptime for info)."));
   kite_add_method(thread, newclass, "format", 
       kite_new_method_compiled_with_docs(thread, hires_format,
                                 "Format date into a string, given format specifier.", 1,
                                 "format", "Format specifier (see man strftime for valid characters)."));
   kite_add_method(thread, newclass, "__construct__",
       kite_new_method_compiled_with_docs(thread, hires_construct, 
                                 "Constructs a new date object, set to epoch (1970 new year's day).", 0));
                                 
   kite_add_method(thread, newclass, "usec",
       kite_new_method_compiled_with_docs(thread, hires_usec, 
                                 "Number of microseconds (0-999999).", 0));    
                                                              
   kite_add_method(thread, newclass, "sec",
       kite_new_method_compiled_with_docs(thread, hires_sec, 
                                 "Number of seconds (0-59).", 0));
   kite_add_method(thread, newclass, "min",
       kite_new_method_compiled_with_docs(thread, hires_min, 
                                 "Number of minutes (0-59).", 0));
   kite_add_method(thread, newclass, "hour",
       kite_new_method_compiled_with_docs(thread, hires_hour, 
                                 "Number of hours (0-23).", 0));
   kite_add_method(thread, newclass, "day_of_month",
       kite_new_method_compiled_with_docs(thread, hires_day_of_month, 
                                 "Day of the month (1-31).", 0));
   kite_add_method(thread, newclass, "month",
       kite_new_method_compiled_with_docs(thread, hires_month, 
                                 "Month of the year (1-12).", 0));
   kite_add_method(thread, newclass, "year",
       kite_new_method_compiled_with_docs(thread, hires_year, 
                                 "Number of seconds (1970-∞).", 0));
                         
  kite_add_property(thread, newclass, "_sec", 0, "Number of seconds since epoch.");
  kite_add_property(thread, newclass, "_usec", 0, "Number of microseconds in the time since epoch.");
  /* Add to parent to complete loading. */
  kite_add_property(thread, parent, "hires", 1, NULL);
  kite_set_property(thread, parent, "hires", newclass);
}