/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

#define TYPECHECK_OPERANDS \
    kite_object_t *rhs; \
    kite_object_t *exc; \
    if (args->type == OBJ_LIST) { \
        rhs = args->builtin_data.listvalue.car; \
    } else { \
        rhs = args; \
    } \
    if (this->type != rhs->type) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Both operands must be of the same type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define TYPECHECK_BOOLEAN_OPERANDS \
    kite_object_t *rhs; \
    kite_object_t *exc; \
    rhs = args; \
    if (this->type != rhs->type) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Both operands must be of the same type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }
  
#define TYPECHECK_RIGHT_OPERAND(t) \
    kite_object_t *rhs; \
    kite_object_t *exc; \
    if (args->type == OBJ_LIST) { \
        rhs = args->builtin_data.listvalue.car; \
    } else { \
        rhs = args; \
    } \
    if (rhs->type != t) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Second operand is not of the requested type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define TYPECHECK_INT_OPERAND TYPECHECK_RIGHT_OPERAND(OBJ_INTEGER)
#define TYPECHECK_METHOD_OPERAND TYPECHECK_RIGHT_OPERAND(OBJ_METHOD)

#define PERFORM_LIST_BOOLEAN_OP(op, c) \
    kite_object_t *obj; \
    TYPECHECK_BOOLEAN_OPERANDS; \
    \
    if (kite_list_count(thd, this) op kite_list_count(thd, rhs) && c != OP_EQUALS) { \
        kite_vm_return(thd, kite_new_boolean(thd, TRUE)); \
        return; \
    } \
    \
    while(this && rhs) { \
        kite_vm_call_operator(thd, this->builtin_data.listvalue.car, \
            c, rhs->builtin_data.listvalue.car, TRUE); \
        obj = kite_vm_pop(thd); \
        \
        if (obj->type == OBJ_BOOLEAN) { \
            if (obj->builtin_data.intvalue == 0) { \
                kite_dereference_object(obj); \
                kite_vm_return(thd, kite_new_boolean(thd, FALSE)); \
                return; \
            } \
            kite_dereference_object(obj); \
        } else { \
            exc = kite_new_exception(thd, "System.exceptions.InvalidArgument", \
                                "Comparison did not result in a boolean."); \
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
            return; \
        } \
        \
        this = this->builtin_data.listvalue.cdr; \
        rhs = rhs->builtin_data.listvalue.cdr; \
    } \
    \
    kite_vm_return(thd, kite_new_boolean(thd, TRUE));

#define CHECK_SINGLE_INDEX \
    index = rhs->builtin_data.intvalue; \
    if (index < 0 || index >= kite_list_count(thd, this)) { \
        exc = kite_new_exception(thd, "System.exceptions.ArrayBounds", \
                            "Provided index is out of bounds."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }
KITE_CLASS_METHOD(List_lshift) {
    kite_object_t *cur = this, *rhs;
    kite_object_t *secondhalf;

    rhs = kite_list_object(thd, args);
    secondhalf = rhs->builtin_data.listvalue.car;

    if (cur->builtin_data.listvalue.car == NULL) {
        cur->builtin_data.listvalue.car = kite_reference_object(secondhalf);
        kite_dereference_object(rhs);
        kite_vm_return(thd, this);
    }

    while(cur->builtin_data.listvalue.cdr != NULL) {
        cur = cur->builtin_data.listvalue.cdr;
    }

    if (secondhalf->type == OBJ_LIST) {
        secondhalf = kite_reference_object(secondhalf);
    } else {
        secondhalf = kite_new_list(thd);
        secondhalf->builtin_data.listvalue.car = 
            kite_reference_object(rhs->builtin_data.listvalue.car);
    }

    cur->builtin_data.listvalue.cdr = secondhalf;
    kite_dereference_object(rhs);
    kite_vm_return(thd, this);
}

KITE_CLASS_METHOD(List_add) {
    kite_object_t *cur = this, *rhs;
    kite_object_t *secondhalf, *new;
    
    new = kite_new_list(thd);
    
    rhs = kite_list_object(thd, args);
    secondhalf = rhs->builtin_data.listvalue.car;
    
    if (secondhalf->type == OBJ_LIST) {
        secondhalf = kite_reference_object(secondhalf);
    } else {
        secondhalf = kite_new_list(thd);
        secondhalf->builtin_data.listvalue.car = 
            kite_reference_object(rhs->builtin_data.listvalue.car);
    }
    
    if (cur->builtin_data.listvalue.car == NULL) {
        if (secondhalf->type == OBJ_LIST)
          new = kite_reference_object(secondhalf);
        else
          new->builtin_data.listvalue.car = kite_reference_object(secondhalf);
        kite_dereference_object(rhs);
        kite_vm_return(thd, new);
        return;
    }
    
    while (cur && cur->builtin_data.listvalue.car) {
          kite_append_list(thd,new,cur->builtin_data.listvalue.car);
          cur = cur->builtin_data.listvalue.cdr;
    }
    cur = new;
    while(cur->builtin_data.listvalue.cdr != NULL) {
        cur = cur->builtin_data.listvalue.cdr;
    }
    
    cur->builtin_data.listvalue.cdr = secondhalf;
    
    kite_dereference_object(rhs);
    kite_vm_return(thd, new);
}

KITE_CLASS_METHOD(List_append) {
    kite_object_t *rhs = args->builtin_data.listvalue.car;
    List_add(thd, this, rhs);
}

KITE_CLASS_METHOD(List_lessthan) {
    PERFORM_LIST_BOOLEAN_OP(<, OP_LESS_THAN);
}

KITE_CLASS_METHOD(List_lessequals) {
    PERFORM_LIST_BOOLEAN_OP(<=, OP_LESS_OR_EQUALS);
}

KITE_CLASS_METHOD(List_equals) {
    PERFORM_LIST_BOOLEAN_OP(==, OP_EQUALS);
}

KITE_CLASS_METHOD(List_greaterequals) {
    PERFORM_LIST_BOOLEAN_OP(>=, OP_GREATER_OR_EQUALS);
}

KITE_CLASS_METHOD(List_greaterthan) {
    PERFORM_LIST_BOOLEAN_OP(>, OP_GREATER_THAN);
}

KITE_CLASS_METHOD(List_notequals) {
    PERFORM_LIST_BOOLEAN_OP(!=, OP_NOT_EQUALS);
}

static void List_derefRange(kite_thread_t *thd, kite_object_t *this, kite_object_t *arg)
{
    kite_object_t *from, *to, *skip, *ret;
    int fromInt, toInt, skipInt, i, j;

    from = kite_get_property(arg, "fromnumber");
    to = kite_get_property(arg, "tonumber");
    skip = kite_get_property(arg, "step");
    fromInt = KITE_GET_INTEGER(from);
    toInt = KITE_GET_INTEGER(to);
    skipInt = KITE_GET_INTEGER(skip);
    kite_dereference_object(from);
    kite_dereference_object(to);
    kite_dereference_object(skip);

    ret = kite_new_list(thd);
    while(this && fromInt > 0)
    {
        fromInt--;
        this = this->builtin_data.listvalue.cdr;
    }

    i = fromInt;
    while(this && i < toInt)
    {
        kite_append_list(thd, ret, this->builtin_data.listvalue.car);
        for(j = 0; j < skipInt && this; j++, i++)
        {
            this = this->builtin_data.listvalue.cdr;
        }
    }

    kite_vm_push(thd, ret);
}

KITE_CLASS_METHOD(List_deref) {
    /* Special-case: return list if range object is given */
    kite_object_t *exc, *rhs;
    int badType = FALSE, index;
    if (args->type == OBJ_LIST) { 
        rhs = args->builtin_data.listvalue.car;
    } else {
        rhs = args;
    }
    if (rhs->type == OBJ_INSTANCE)
    {
        kite_object_t *obj_name = kite_object_name(thd, rhs);
        if (!strcmp("System.math.range", KITE_GET_STRING_VALUE(obj_name)))
        {
            List_derefRange(thd, this, rhs);
            kite_dereference_object(obj_name);
            return;
        }
        else
        {
            badType = TRUE;
            kite_dereference_object(obj_name);
        }
    }
    else if (rhs->type != OBJ_INTEGER)
    {
        badType = TRUE;
    }

    if (badType == TRUE)
    {
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch",
                            "Second operand is not of the requested type.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }

    CHECK_SINGLE_INDEX;
    
    while(index > 0) {
        index--;
        this = this->builtin_data.listvalue.cdr;
    }

    kite_vm_return(thd, KITE_LIST_CAR(this));
}

KITE_CLASS_METHOD(List_set) {
    /* Special-case: return list if range object is given */
    kite_object_t *exc, *rhs, *val;
    int index;

    KITE_GET_METHOD_ARGUMENT(rhs, 1);
    KITE_GET_METHOD_ARGUMENT(val, 2);

    CHECK_SINGLE_INDEX;
    
    while(index > 0) {
        index--;
        this = this->builtin_data.listvalue.cdr;
    }

    KITE_LIST_CAR(this) = kite_reference_object(val);
}

KITE_CLASS_METHOD(List_str) {
    kite_object_t *cur = this;
    kite_object_t *stringrep;
    char *ret = strdup("[");
    
    KITE_NO_ARGS;

    while(cur && cur->builtin_data.listvalue.car) {
        stringrep = kite_string_object(thd, cur->builtin_data.listvalue.car);
        ret = realloc(ret, strlen(ret) + 
            strlen(stringrep->builtin_data.stringvalue.string) + 3);
        strcat(ret, stringrep->builtin_data.stringvalue.string);
        kite_dereference_object(stringrep);
        if (cur->builtin_data.listvalue.cdr) {
            strcat(ret, ", ");
        }
        cur = cur->builtin_data.listvalue.cdr;
    }
    
    strcat(ret, "]");
    stringrep = kite_new_string(thd, ret);
    free(ret);
    kite_vm_return(thd, stringrep);
}

KITE_CLASS_METHOD(List_count) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_new_integer(thd, kite_list_count(thd, this)));
}

KITE_CLASS_METHOD(List_head) {
    kite_object_t *ret = this->builtin_data.listvalue.car;

    KITE_NO_ARGS;

    if (ret == NULL) ret = kite_new_null(thd);
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(List_tail) {
    kite_object_t *ret = this->builtin_data.listvalue.cdr;
    KITE_NO_ARGS;
    if (ret == NULL) ret = kite_new_list(thd);
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(List_reset) {
    KITE_NO_ARGS;
    (void)thd;
    
    if (this->builtin_data.listvalue.iterator) {
        kite_dereference_object(this->builtin_data.listvalue.iterator);
    }
    this->builtin_data.listvalue.iterator = kite_reference_object(this);
}

KITE_CLASS_METHOD(List_cur) {
    kite_object_t *pos = this->builtin_data.listvalue.iterator;
    KITE_NO_ARGS;
    
    if (pos == NULL || pos->builtin_data.listvalue.car == NULL) {
        kite_vm_return(thd, kite_new_null(thd));
    }
    
    kite_vm_return(thd, pos->builtin_data.listvalue.car);
}

KITE_CLASS_METHOD(List_next) {
    KITE_NO_ARGS;
    
    if (this->builtin_data.listvalue.iterator == NULL || 
        this->builtin_data.listvalue.iterator->builtin_data.listvalue.car == NULL) {
        kite_vm_return(thd, kite_new_boolean(thd, FALSE));
    } else {
        kite_object_t *tmp = this->builtin_data.listvalue.iterator;
        this->builtin_data.listvalue.iterator = 
            kite_reference_object(tmp->builtin_data.listvalue.cdr);
        kite_dereference_object(tmp);
        if (this->builtin_data.listvalue.iterator == NULL || 
            this->builtin_data.listvalue.iterator->builtin_data.listvalue.car == NULL) {
            kite_vm_return(thd, kite_new_boolean(thd, FALSE));
        } else {
            (void)kite_reference_object(this);
            kite_vm_return(thd, kite_new_boolean(thd, TRUE));
        }
    }
}

KITE_CLASS_METHOD(List_map) {
    kite_object_t *ret, *cur = this, *curret, *methodargs;
    TYPECHECK_METHOD_OPERAND;
    
    ret = curret = kite_new_list(thd);
    while(cur) {
        methodargs = kite_list_object(thd, cur->builtin_data.listvalue.car);
        kite_vm_call_object(thd, this, rhs, methodargs);
        curret->builtin_data.listvalue.car = kite_vm_pop(thd);
        if (cur->builtin_data.listvalue.cdr) {
            curret->builtin_data.listvalue.cdr = kite_new_list(thd);
        }
        
        kite_dereference_object(methodargs);
        cur = cur->builtin_data.listvalue.cdr;
        curret = curret->builtin_data.listvalue.cdr;
    }
    
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(List_reduce) {
    kite_object_t *ret, *cur = this, *methodargs;
    TYPECHECK_METHOD_OPERAND;
    
    ret = kite_new_null(thd);
    while(cur) {
        methodargs = kite_new_list(thd);
        methodargs->builtin_data.listvalue.car = kite_reference_object(ret);
        methodargs->builtin_data.listvalue.cdr = 
            kite_list_object(thd, cur->builtin_data.listvalue.car);
        if (rhs->builtin_data.funcvalue.this)
            kite_vm_call_object(thd, rhs->builtin_data.funcvalue.this, rhs, methodargs);
        else
            kite_vm_call_object(thd, this, rhs, methodargs);
        kite_dereference_object(ret);
        ret = kite_vm_pop(thd);
        kite_dereference_object(methodargs);
        cur = cur->builtin_data.listvalue.cdr;
    }
    
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(List_removeAt) {
    int index;
    kite_object_t *cur = this, *desired, *tail;

    TYPECHECK_INT_OPERAND;
    CHECK_SINGLE_INDEX;
    
    while(index > 1) {
        index--;
        cur = cur->builtin_data.listvalue.cdr;
    }
    
    desired = cur->builtin_data.listvalue.cdr;
    if (index == 0) desired = this;
    tail = desired->builtin_data.listvalue.cdr;
    if (tail) {
        kite_dereference_object(cur->builtin_data.listvalue.car);
        desired->builtin_data.listvalue.car = 
            kite_reference_object(tail->builtin_data.listvalue.car);
        desired->builtin_data.listvalue.cdr = 
            kite_reference_object(tail->builtin_data.listvalue.cdr);
        kite_dereference_object(tail);
    } else {
        if (desired != this) kite_dereference_object(desired);
        else {
            kite_dereference_object(desired->builtin_data.listvalue.car);
            desired->builtin_data.listvalue.car = NULL;
        }
        cur->builtin_data.listvalue.cdr = NULL;
    }
    kite_vm_return(thd, this);
}

KITE_CLASS_METHOD(List_prepend) {
    kite_object_t *arg = args->builtin_data.listvalue.car;
    kite_object_t *ret;
    
    if (this->builtin_data.listvalue.car) {
        ret = kite_new_list(thd);
        ret->builtin_data.listvalue.car = kite_reference_object(this->builtin_data.listvalue.car);
        ret->builtin_data.listvalue.cdr = kite_reference_object(this->builtin_data.listvalue.cdr);
        kite_dereference_object(this->builtin_data.listvalue.car);
        kite_dereference_object(this->builtin_data.listvalue.cdr);
        this->builtin_data.listvalue.car = kite_reference_object(arg);
        this->builtin_data.listvalue.cdr = ret;
    } else {
        this->builtin_data.listvalue.car = kite_reference_object(arg);
    }
    kite_vm_return(thd, this);
}

KITE_CLASS_METHOD(List_sublist) {
    int index;
    int length;
    kite_object_t *ret, *cur = this, *curret;

    TYPECHECK_INT_OPERAND;
    CHECK_SINGLE_INDEX;
    
    if (args->builtin_data.listvalue.cdr) {
        TYPECHECK_INT_OPERAND;
        args = args->builtin_data.listvalue.cdr;
        length = rhs->builtin_data.intvalue;
    } else {
        length = INT_MAX;
    }
    
    /* Skip beginning */
    while(index > 0) {
        index--;
        cur = cur->builtin_data.listvalue.cdr;
    }
    
    /* Create new list object */
    ret = kite_new_list(thd);
    curret = ret;
    while(cur && length > 0) {
        curret->builtin_data.listvalue.car = 
            kite_reference_object(cur->builtin_data.listvalue.car);
        if (cur->builtin_data.listvalue.cdr && length > 1) {
            curret->builtin_data.listvalue.cdr = kite_new_list(thd);
        }
        cur = cur->builtin_data.listvalue.cdr;
        curret = curret->builtin_data.listvalue.cdr;
        length--;
    }
    
    kite_vm_return(thd, ret);
}

void kite_concat_list(kite_thread_t *thd, kite_object_t *list, kite_object_t *otherList) {
  kite_object_t *cur, *rest;
  if (otherList->type == OBJ_LIST) {
    cur = KITE_LIST_CAR(otherList);
    rest = KITE_LIST_CDR(otherList);
    while (cur != NULL) {
      kite_append_list(thd,list,cur);
      if (rest != NULL)
        rest = KITE_LIST_CDR(rest);
      if (rest != NULL) {
        cur = KITE_LIST_CAR(rest);
      }
      else
        cur = NULL;
    }
  } else {
    kite_append_list(thd,list,otherList);
  }
}

static kite_object_t* sortKiteList(kite_thread_t *thd, kite_object_t *list, kite_object_t *method) {  
    kite_object_t *partition, *rest, *left, *right, *cur;
  
  if (KITE_LIST_CDR(list) == NULL) return list;
  partition = KITE_LIST_CAR(list);
  rest      = KITE_LIST_CDR(list);

  left = kite_new_list(thd);
  right = kite_new_list(thd);

  cur = KITE_LIST_CAR(rest);
  
  while (cur != NULL) {
    kite_object_t *args = kite_new_list(thd);
    long result;
    
    kite_append_list(thd,args,partition);
    kite_append_list(thd,args,cur);
    
    kite_vm_call_object(thd, kite_new_null(thd), method, args);
    result = KITE_GET_INTEGER(kite_vm_pop(thd));
    
    switch (result) {
      case -1:
        kite_append_list(thd,left,cur);
        break;
      case 0:
        kite_append_list(thd,right,cur);
        break;
      case 1:
        kite_append_list(thd,right,cur);
        break;
    }
    rest = KITE_LIST_CDR(rest);
    if (rest != NULL) {
      cur = KITE_LIST_CAR(rest);
    }
    else
      cur = NULL;
    
    kite_dereference_object(args);
    
  }
  
  left = sortKiteList(thd,left,method);
  right = sortKiteList(thd,right,method);
  kite_append_list(thd,left,partition);
  kite_concat_list(thd,left,right);
  
  kite_dereference_object(right);
  kite_dereference_object(list);
  
  return left;
}

KITE_CLASS_METHOD(List_sort)
{
  //grab the file argument
  kite_object_t *method, *result;
  
  KITE_GET_METHOD_ARGUMENT(method,1);
  
  result = sortKiteList(thd,this,method);
  
  kite_vm_return(thd, result);
}

KITE_MODULE_INITIALIZER(list) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "list");
    kite_set_docstring(newclass, "Object representing a list of objects.");
    
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "list", 1, NULL);
    kite_set_property(thread, parent, "list", newclass);
    
    /*
     * Add operators
     */
    kite_add_operator(thread, newclass, OP_ADD, 
        kite_new_method_compiled_with_docs(thread, List_add, 
        "Creates a new list that is the combination of the two operands.", 1,
        "item", "Item to add."));

   kite_add_operator(thread, newclass, OP_LEFT_SHIFT, 
       kite_new_method_compiled_with_docs(thread, List_lshift, 
                                 "Adds item to the end of the list.", 1,
                                 "item", "Item to add."));
                                 
    kite_add_operator(thread, newclass, OP_LESS_THAN, 
        kite_new_method_compiled_with_docs(thread, List_lessthan, 
                                  "Returns true if every item in the left list is less than every item in the right.", 1,
                                  "rhs", "The list to compare to."));
    kite_add_operator(thread, newclass, OP_LESS_OR_EQUALS, 
        kite_new_method_compiled_with_docs(thread, List_lessequals,
                                  "Returns true if every item in the left list is less than or eaual to every item in the right.", 1,
                                  "rhs", "The list to compare to."));
    kite_add_operator(thread, newclass, OP_EQUALS, 
        kite_new_method_compiled_with_docs(thread, List_equals, 
                                  "Compares two lists for equality.", 1,
                                  "rhs", "The list to compare to."));
    kite_add_operator(thread, newclass, OP_GREATER_OR_EQUALS, 
        kite_new_method_compiled_with_docs(thread, List_greaterequals,
                                  "Returns true if every item in the left list is greater than or equal to every item in the right.", 1,
                                  "rhs", "The list to compare to."));
    kite_add_operator(thread, newclass, OP_GREATER_THAN, 
        kite_new_method_compiled_with_docs(thread, List_greaterthan,
                                  "Returns true if every item in the left list is greater than every item in the right.", 1,
                                  "rhs", "The list to compare to."));
    kite_add_operator(thread, newclass, OP_NOT_EQUALS, 
        kite_new_method_compiled_with_docs(thread, List_notequals, 
                                  "Compares two lists for lack of equality.", 1,
                                  "rhs", "The list to compare to."));
    kite_add_operator(thread, newclass, OP_ARRAY_DEREF, 
        kite_new_method_compiled_with_docs(thread, List_deref, 
                                  "Returns the item at the specified index.", 1,
                                  "index", "The index of the item to return."));
    kite_add_operator(thread, newclass, OP_ARRAY_SET, 
        kite_new_method_compiled_with_docs(thread, List_set, 
                                  "Set the item at the specified index.", 2,
                                  "index", "The index of the item to return.",
                                  "value", "The value to set index to."));
    kite_add_operator(thread, newclass, OP_MAP, 
        kite_new_method_compiled_with_docs(thread, List_map, 
                                  "Calls method on each of the elements in the list and returns a new list.", 1,
                                  "method", "Method to call. It should accept one argument (the current item)."));
    kite_add_operator(thread, newclass, OP_REDUCE, 
        kite_new_method_compiled_with_docs(thread, List_reduce, 
                                  "Calls method on each of the elements and returns a single result.", 1,
                                  "method", "Method to call. It should accept two arguments (the current result and the current item)."));
    
    /* Add methods. */
    kite_add_method(thread, newclass, "str", 
        kite_new_method_compiled_with_docs(thread, List_str, 
                                  "Returns string representaiton of the list.", 0));
    kite_add_method(thread, newclass, "reset", 
        kite_new_method_compiled_with_docs(thread, List_reset, 
                                  "Resets list's iterator to the first element.", 0));
    kite_add_method(thread, newclass, "cur", 
        kite_new_method_compiled_with_docs(thread, List_cur, 
                                  "Returns current item pointed to by the list's iterator.", 0));
    kite_add_method(thread, newclass, "next", 
        kite_new_method_compiled_with_docs(thread, List_next, 
                                  "Moves iterator to the next item in the list.", 0));
    kite_add_method(thread, newclass, "count", 
        kite_new_method_compiled_with_docs(thread, List_count, 
                                  "Returns the number of items in the list.", 0));
    kite_add_method(thread, newclass, "head", 
        kite_new_method_compiled_with_docs(thread, List_head, 
                                  "Returns the first item in the list.", 0));
    kite_add_method(thread, newclass, "tail", 
        kite_new_method_compiled_with_docs(thread, List_tail, 
                                  "Returns a list containing all but the first item.", 0));
    kite_add_method(thread, newclass, "append", 
        kite_new_method_compiled_with_docs(thread, List_append, 
                                  "Adds an item to the end of the list.", 1,
                                  "item", "Item to add."));
    kite_add_method(thread, newclass, "getIndex", 
        kite_new_method_compiled_with_docs(thread, List_deref, 
                                  "Returns the item at the specified index.", 1,
                                  "index", "The 0-based index of the item to return."));
    kite_add_method(thread, newclass, "prepend", 
        kite_new_method_compiled_with_docs(thread, List_prepend, 
                                  "Adds an item to the beginning of the list.", 1,
                                  "item", "Item to add."));
    kite_add_method(thread, newclass, "sublist", 
        kite_new_method_compiled_with_docs(thread, List_sublist, 
                                  "Returns the list containing every item from the specified index onwards.", 1,
                                  "index", "The 0-based index of the first item to return."));
    kite_add_method(thread, newclass, "sublist", 
        kite_new_method_compiled_with_docs(thread, List_sublist,
                                  "Returns the list containing every item from the specified range of indices.", 2,
                                  "from", "The 0-based index of the first item to return.",
                                  "count", "The number of items to return."));
    kite_add_method(thread, newclass, "removeAt", 
        kite_new_method_compiled_with_docs(thread, List_removeAt, 
                                  "Removes the item at the specified index.", 1,
                                  "index", "The 0-based index of the item to remove."));
    kite_add_method(thread, newclass, "sort",
        kite_new_method_compiled_with_docs(thread, List_sort,
                                  "Sort a list of objects.", 1,
                                  "comparator","The method used to sort the objects."));
}

