/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#define _USE_MATH_DEFINES
#include <math.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

#define TYPECHECK_RIGHT_OPERAND(t) \
    kite_object_t *rhs = args->builtin_data.listvalue.car; \
    kite_object_t *exc; \
    if (rhs->type != t) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Second operand is not of the requested type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define TYPECHECK_RIGHT_OPERANDS(t) \
    kite_object_t *rhs1 = args->builtin_data.listvalue.car; \
    kite_object_t *rhs2 = \
        args->builtin_data.listvalue.cdr->builtin_data.listvalue.car; \
    kite_object_t *exc; \
    if (rhs1->type != t || rhs2->type != t) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Operand(s) are not of the requested type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }
    
#define TYPECHECK_INT_OPERAND TYPECHECK_RIGHT_OPERAND(OBJ_INTEGER)
#define TYPECHECK_FLOAT_OPERAND TYPECHECK_RIGHT_OPERAND(OBJ_FLOAT)
#define TYPECHECK_FLOAT_OPERANDS TYPECHECK_RIGHT_OPERANDS(OBJ_FLOAT)

#define MATH_H_SINGLE_FLOAT_PARAM(name) \
    KITE_CLASS_METHOD(Math_ ## name) { \
        TYPECHECK_FLOAT_OPERAND; \
        (void)this; \
        kite_vm_return(thd, \
            kite_new_float(thd, name(rhs->builtin_data.floatvalue))); \
    }

#define MATH_H_SINGLE_FLOAT_PARAM_RETURN_INT(name) \
    KITE_CLASS_METHOD(Math_ ## name) { \
        TYPECHECK_FLOAT_OPERAND; \
        (void)this; \
        kite_vm_return(thd, \
            kite_new_integer(thd, name(rhs->builtin_data.floatvalue))); \
    }

#define MATH_H_NO_PARAM_RETURN_INT(name) \
    KITE_CLASS_METHOD(Math_ ## name) { \
        (void)this; \
        (void)args; \
        kite_vm_return(thd, \
            kite_new_integer(thd, name())); \
    }
    
#define MATH_H_SINGLE_INT_PARAM(name) \
    KITE_CLASS_METHOD(Math_ ## name) { \
        TYPECHECK_INT_OPERAND; \
        (void)this; \
        name(rhs->builtin_data.intvalue); \
    }
    
#define MATH_H_TWO_FLOAT_PARAMS(name) \
    KITE_CLASS_METHOD(Math_ ## name) { \
        TYPECHECK_FLOAT_OPERANDS; \
        (void)this; \
        kite_vm_return(thd, \
            kite_new_float(thd, name(rhs1->builtin_data.floatvalue, \
            rhs2->builtin_data.floatvalue))); \
    }

#define METHOD_ADD_0(name, param, doc) \
    kite_add_method(thread, newclass, # name, \
        kite_new_method_compiled_with_docs(thread, Math_ ## name, doc, param));

#define METHOD_ADD_1(name, param, doc, param1, desc1) \
    kite_add_method(thread, newclass, # name, \
        kite_new_method_compiled_with_docs(thread, Math_ ## name, doc, param, param1, desc1));

#define METHOD_ADD_2(name, param, doc, param1, desc1, param2, desc2) \
    kite_add_method(thread, newclass, # name, \
        kite_new_method_compiled_with_docs(thread, Math_ ## name, doc, param, param1, desc1, param2, desc2));

#define CONSTANT_ADD(name, value) \
    kite_add_property(thread, newclass, # name, 1, # name); \
    kite_set_property(thread, newclass, # name, kite_new_float(thread, value));
    
MATH_H_SINGLE_FLOAT_PARAM(acos);
MATH_H_SINGLE_FLOAT_PARAM(asin);
MATH_H_SINGLE_FLOAT_PARAM(atan);
MATH_H_TWO_FLOAT_PARAMS(atan2);
MATH_H_SINGLE_FLOAT_PARAM(ceil);
MATH_H_SINGLE_FLOAT_PARAM(cos);
MATH_H_SINGLE_FLOAT_PARAM(cosh);
MATH_H_SINGLE_FLOAT_PARAM(exp);
MATH_H_SINGLE_FLOAT_PARAM(fabs);
MATH_H_SINGLE_FLOAT_PARAM(floor);
MATH_H_TWO_FLOAT_PARAMS(fmod);
//frexp
//ldexp
MATH_H_SINGLE_FLOAT_PARAM(log);
MATH_H_SINGLE_FLOAT_PARAM(log10);
//modf
MATH_H_TWO_FLOAT_PARAMS(pow);
MATH_H_SINGLE_FLOAT_PARAM(sin);
MATH_H_SINGLE_FLOAT_PARAM(sinh);
MATH_H_SINGLE_FLOAT_PARAM(sqrt);
MATH_H_SINGLE_FLOAT_PARAM(tan);
MATH_H_SINGLE_FLOAT_PARAM(tanh);

#ifndef WIN32
MATH_H_SINGLE_FLOAT_PARAM(erf);
MATH_H_SINGLE_FLOAT_PARAM(erfc);
MATH_H_SINGLE_FLOAT_PARAM(lgamma);
MATH_H_SINGLE_FLOAT_PARAM(acosh);
MATH_H_SINGLE_FLOAT_PARAM(asinh);
MATH_H_SINGLE_FLOAT_PARAM(atanh);
MATH_H_SINGLE_FLOAT_PARAM_RETURN_INT(isnan);
MATH_H_SINGLE_FLOAT_PARAM(cbrt);
MATH_H_SINGLE_FLOAT_PARAM_RETURN_INT(ilogb);
MATH_H_SINGLE_FLOAT_PARAM(logb);
MATH_H_TWO_FLOAT_PARAMS(nextafter);
MATH_H_TWO_FLOAT_PARAMS(remainder);
MATH_H_SINGLE_FLOAT_PARAM(rint);
MATH_H_TWO_FLOAT_PARAMS(scalb);
#endif /* WIN32 */

MATH_H_TWO_FLOAT_PARAMS(hypot);
MATH_H_SINGLE_FLOAT_PARAM(j0);
MATH_H_SINGLE_FLOAT_PARAM(j1);
//jn

MATH_H_SINGLE_FLOAT_PARAM(y0);
MATH_H_SINGLE_FLOAT_PARAM(y1);
//yn

KITE_CLASS_METHOD(Math_random)
{
    (void)this;
    (void)args;
#ifdef WIN32
    kite_vm_push(thd, kite_new_integer(thd, rand()));
#else
    kite_vm_push(thd, kite_new_integer(thd, random()));
#endif /* WIN32 */
}

KITE_CLASS_METHOD(Math_srandom) {
    TYPECHECK_INT_OPERAND;
    (void)this;
#ifndef WIN32
    srandom(rhs->builtin_data.intvalue);
#else
    srand(rhs->builtin_data.intvalue);
#endif /* WIN32 */
}

KITE_MODULE_INITIALIZER(math) {
    /* Create new class object. inherits is typically NULL for packages. */
    kite_object_t *newclass = kite_new_class(thread, NULL, "math");
    kite_set_docstring(newclass, "Mathematical functions and constants.");
    
    /* Add methods common to all classes. */
    METHOD_ADD_1(acos, 1, "Returns the arccosine of x.", "x", "The input to the function.");
    METHOD_ADD_1(asin, 1, "Returns the arcsine of x.", "x", "The input to the function.");
    METHOD_ADD_1(atan, 1, "Returns the arctangent of x.", "x", "The input to the function.");
    METHOD_ADD_2(atan2, 2, "Returns the arctangent of y/x.", "x", "The first input to the function.", "y", "The second input to the function.");
    METHOD_ADD_1(ceil, 1, "Returns the smallest integer not less than x.", "x", "The input to the function.");
    METHOD_ADD_1(cos, 1, "Returns the cosine of x.", "x", "The input to the function.");
    METHOD_ADD_1(cosh, 1, "Returns the hyperbolic cosine of x.", "x", "The input to the function.");
    METHOD_ADD_1(exp, 1, "Returns the value of e raised to x.", "x", "The input to the function.");
    METHOD_ADD_1(fabs, 1, "Returns the absolute value of x.", "x", "The input to the function.");
    METHOD_ADD_1(floor, 1, "Returns the largest integer not greater than x.", "x", "The input to the function.");
    METHOD_ADD_2(fmod, 2, "Returns the floating-point remainder of x/y.", "x", "The first input to the function.", "y", "The second input to the function.");
    //METHOD_ADD(frexp, 1);
    //METHOD_ADD(ldexp, 2);
    METHOD_ADD_1(log, 1, "Returns the natural logarithm of x.", "x", "The input to the function.");
    METHOD_ADD_1(log10, 1, "Returns the logarithm base 10 of x.", "x", "The input to the function.");
    //METHOD_ADD(modf, 1);
    METHOD_ADD_2(pow, 2, "Returns x raised to the power of y.", "x", "The first input to the function.", "y", "The second input to the function.");
    METHOD_ADD_1(sin, 1, "Returns the sine of x.", "x", "The input to the function.");
    METHOD_ADD_1(sinh, 1, "Returns the hyperbolic sine of x.", "x", "The input to the function.");
    METHOD_ADD_1(sqrt, 1, "Returns the square root of x.", "x", "The input to the function.");
    METHOD_ADD_1(tan, 1, "Returns the tangent of x.", "x", "The input to the function.");
    METHOD_ADD_1(tanh, 1, "Returns the hyperbolic tangent of x.", "x", "The input to the function.");

    METHOD_ADD_2(hypot, 2, "Returns sqrt(x*x + y*y).", "x", "The first input to the function.", "y", "The second input to the function.");
    METHOD_ADD_1(j0, 1, "Returns the Bessel function of the first kind (order 0) of x.", "x", "The input to the function.");
    METHOD_ADD_1(j1, 1, "Returns the Bessel function of the first kind (order 1) of x.", "x", "The input to the function.");
    //METHOD_ADD(jn, 2);
    METHOD_ADD_1(y0, 1, "Returns the Bessel function of the second kind (order 0) of x.", "x", "The input to the function.");
    METHOD_ADD_1(y1, 1, "Returns the Bessel function of the second kind (order 1) of x.", "x", "The input to the function.");
    //METHOD_ADD(yn, 2);

    METHOD_ADD_0(random, 0, "Returns a random number");
    METHOD_ADD_1(srandom, 1, "Sets random number generator's seed.", "x", "Seed to use.");

#ifndef WIN32
    METHOD_ADD_1(erf, 1, "Returns the error function of x.", "x", "The input to the function.");
    METHOD_ADD_1(erfc, 1, "Returns the complimentary error function of x.", "x", "The input to the function.");
    METHOD_ADD_1(lgamma, 1, "Returns the log_e of gamma(|x|).", "x", "The input to the function.");
    METHOD_ADD_1(isnan, 1, "Returns whether the given number is NaN.", "x", "The input to the function.");
    METHOD_ADD_1(acosh, 1, "Returns the inverse hyperbolic cosine of x.", "x", "The input to the function.");
    METHOD_ADD_1(asinh, 1, "Returns the inverse hyperbolic sine of x.", "x", "The input to the function.");
    METHOD_ADD_1(atanh, 1, "Returns the inverse hyperbolic tangent of x.", "x", "The input to the function.");
    METHOD_ADD_1(cbrt, 1, "Returns the cubic root of x.", "x", "The input to the function.");
    METHOD_ADD_1(ilogb, 1, "Returns the exponent of x (as integer).", "x", "The input to the function.");
    METHOD_ADD_1(logb, 1, "Returns the exponent of x (as floating-point).", "x", "The input to the function.");
    METHOD_ADD_2(nextafter, 2, "Returns the next representable number from x in direction y.", "x", "The first input to the function.", "y", "The second input to the function.");
    METHOD_ADD_2(remainder, 2, "Returns the remainder of x/y.", "x", "The first input to the function.", "y", "The second input to the function.");
    METHOD_ADD_1(rint, 1, "Returns x rounded to an integer.", "x", "The input to the function.");
    METHOD_ADD_2(scalb, 2, "Returns the value of x*r^y.", "x", "The first input to the function.", "y", "The second input to the function.");
#endif /* WIN32 */

    /* Add constants */
    CONSTANT_ADD(E, M_E);
    CONSTANT_ADD(LOG2E, M_LOG2E);
    CONSTANT_ADD(LOG10E, M_LOG10E);
    CONSTANT_ADD(LN2, M_LN2);
    CONSTANT_ADD(LN10, M_LN10);
    CONSTANT_ADD(PI, M_PI);
    CONSTANT_ADD(PI_2, M_PI_2);
    CONSTANT_ADD(PI_4, M_PI_4);
    CONSTANT_ADD(ONE_PI, M_1_PI);
    CONSTANT_ADD(TWO_PI, M_2_PI);
    CONSTANT_ADD(TWO_SQRTPI, M_2_SQRTPI);
    CONSTANT_ADD(SQRT2, M_SQRT2);
    CONSTANT_ADD(SQRT1_2, M_SQRT1_2);
    
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "math", 1, NULL);
    kite_set_property(thread, parent, "math", newclass);
}

