/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

/*****************************************************************************
 * Reset iterator
 ****************************************************************************/
KITE_CLASS_METHOD(Range_reset) {
    kite_symtab_t *symt;
    kite_object_t *from, *cur;
    
    KITE_NO_ARGS;
    (void)thd;
    
    KITE_FIND_PROPERTY(symt, this, "fromnumber", 0);
    from = symt->value;
    KITE_FIND_PROPERTY(symt, this, "current", 0);
    cur = symt->value;
    
    memcpy(&cur->builtin_data, &from->builtin_data, 
           sizeof(from->builtin_data));
}

/*****************************************************************************
 * Return current number in range
 ****************************************************************************/
KITE_CLASS_METHOD(Range_cur) {
    kite_symtab_t *symt;
    kite_object_t *cur, *max;
    
    KITE_NO_ARGS;
    
    KITE_FIND_PROPERTY(symt, this, "current", 0);
    cur = symt->value;
    KITE_FIND_PROPERTY(symt, this, "tonumber", 0);
    max = symt->value;
    
    switch(cur->type) {
        case OBJ_FLOAT:
        {
            if (cur->builtin_data.floatvalue >= max->builtin_data.floatvalue)
            {
                kite_vm_return(thd, kite_new_null(thd));
            }
            kite_vm_return(thd, cur);
            break;
        }
        case OBJ_INTEGER:
        {
            if (cur->builtin_data.intvalue >= max->builtin_data.intvalue)
            {
                kite_vm_return(thd, kite_new_null(thd));
            }
            kite_vm_return(thd, cur);
            break;
        }
        default:
        {
            assert(0);
        }
    }
}

/*****************************************************************************
 * Increment current number in range
 ****************************************************************************/
KITE_CLASS_METHOD(Range_next) {
    kite_symtab_t *symt;
    kite_object_t *cur, *max, *step;
    
    KITE_NO_ARGS;
    
    KITE_FIND_PROPERTY(symt, this, "current", 0);
    cur = symt->value;
    KITE_FIND_PROPERTY(symt, this, "tonumber", 0);
    max = symt->value;
    KITE_FIND_PROPERTY(symt, this, "step", 0);
    step = symt->value;
    
    switch(cur->type) {
        case OBJ_FLOAT:
        {
            cur->builtin_data.floatvalue += step->builtin_data.floatvalue;
            if (cur->builtin_data.floatvalue >= max->builtin_data.floatvalue)
            {
                kite_vm_return(thd, kite_new_boolean(thd, FALSE));
            }
            kite_vm_return(thd, kite_new_boolean(thd, TRUE));
            break;
        }
        case OBJ_INTEGER:
        {
            cur->builtin_data.intvalue += step->builtin_data.intvalue;
            if (cur->builtin_data.intvalue >= max->builtin_data.intvalue)
            {
                kite_vm_return(thd, kite_new_boolean(thd, FALSE));
            }
            kite_vm_return(thd, kite_new_boolean(thd, TRUE));
            break;
        }
        default:
        {
            assert(0);
        }
    }
}

/*****************************************************************************
 * Construct range object
 ****************************************************************************/
KITE_CLASS_METHOD(Range_construct) {
    kite_object_t *from, *to, *step = NULL, *second, *tmp, *current;
    
    from = args->builtin_data.listvalue.car;
    second = args->builtin_data.listvalue.cdr;
    to = second->builtin_data.listvalue.car;
    if (kite_list_count(thd, args) > 2) {
        second = second->builtin_data.listvalue.cdr;
        step = second->builtin_data.listvalue.car;
    }
    
    switch(from->type) {
        case OBJ_INTEGER:
        {
            from = kite_new_integer(thd, from->builtin_data.intvalue);
            tmp = kite_int_object(thd, to);
            to = kite_new_integer(thd, tmp->builtin_data.intvalue);
            kite_dereference_object(tmp);
            if (step != NULL) {
                tmp = kite_int_object(thd, step);
                step = kite_new_integer(thd, tmp->builtin_data.intvalue);
                kite_dereference_object(tmp);
            } else {
                step = kite_new_integer(thd, 1);
            }
            current = kite_new_integer(thd, from->builtin_data.intvalue);
            break;
        }
        case OBJ_FLOAT:
        {
            from = kite_new_float(thd, from->builtin_data.floatvalue);
            tmp = kite_float_object(thd, to);
            to = kite_new_float(thd, tmp->builtin_data.floatvalue);
            kite_dereference_object(tmp);
            if (step != NULL) {
                tmp = kite_float_object(thd, step);
                step = kite_new_float(thd, tmp->builtin_data.floatvalue);
                kite_dereference_object(tmp);
            } else {
                step = kite_new_float(thd, 1.0);
            }
            current = kite_new_float(thd, from->builtin_data.floatvalue);
            break;
        }
        default:
        {
            kite_object_t *exc;
            exc = kite_new_exception(thd, "System.exceptions.TypeMismatch",
                                     "Only integers and floats are supported.");
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
            return;
        }
    }
    
    kite_set_property(thd, this, "fromnumber", from);
    kite_set_property(thd, this, "tonumber", to);
    kite_set_property(thd, this, "step", step);
    kite_set_property(thd, this, "current", current);
}

KITE_MODULE_INITIALIZER(range) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "range");
    kite_set_docstring(newclass, "Object representing a range of numbers.");
    
    /* Add methods. */
    kite_add_method(thread, newclass, "reset", 
        kite_new_method_compiled_with_docs(thread, Range_reset, 
                                  "Resets range to the beginning.", 0));
    kite_add_method(thread, newclass, "cur", 
        kite_new_method_compiled_with_docs(thread, Range_cur, 
                                  "Returns the current number in the range.", 0));
    kite_add_method(thread, newclass, "next", 
        kite_new_method_compiled_with_docs(thread, Range_next, 
                                  "Moves pointer to next object in range.", 0));
    kite_add_method(thread, newclass, "__construct__", 
        kite_new_method_compiled_with_docs(thread, Range_construct, 
                                  "Construct new range object.", 2,
                                  "from", "Number to start with.",
                                  "to", "Number to end with."));
    kite_add_method(thread, newclass, "__construct__", 
        kite_new_method_compiled_with_docs(thread, Range_construct,
                                  "Construct new range object.", 3,
                                  "from", "Number to start with.",
                                  "to", "Number to end with.",
                                  "step", "Number to skip between elements in range."));

    /* Add common properties. */
    kite_add_property(thread, newclass, "fromnumber", 0, "The beginning of the given range.");
    kite_add_property(thread, newclass, "tonumber", 0, "The end of the given range.");
    kite_add_property(thread, newclass, "step", 0, "The amount to skip between numbers in the range.");
    kite_add_property(thread, newclass, "current", 0, "Current value in range.");
    
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "range", 1, NULL);
    kite_set_property(thread, parent, "range", newclass);
}
