/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include "kite_object.h"
#include "vm/kite_vm.h"

KITE_CLASS_METHOD(Null_str) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_string_object(thd, this));
}

KITE_CLASS_METHOD(Null_int) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_int_object(thd, this));
}

KITE_CLASS_METHOD(Null_float) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_float_object(thd, this));
}

KITE_CLASS_METHOD(Null_bool) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_boolean_object(thd, this));
}

KITE_MODULE_INITIALIZER(null) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "null");
    kite_set_docstring(newclass, "An object representing nothing.");
    
    /* Add methods. */
    kite_add_method(thread, newclass, "str", 
        kite_new_method_compiled_with_docs(thread, Null_str, 
                                  "Return string representation of object.", 0));
    kite_add_method(thread, newclass, "int", 
        kite_new_method_compiled_with_docs(thread, Null_int, 
                                  "Return integer representation of object.", 0));
    kite_add_method(thread, newclass, "float", 
        kite_new_method_compiled_with_docs(thread, Null_float, 
                                  "Return floating-point representation of object.", 0));
    kite_add_method(thread, newclass, "bool", 
        kite_new_method_compiled_with_docs(thread, Null_bool, 
                                  "Return Boolean representation of object.", 0));

    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "null", 1, NULL);
    kite_set_property(thread, parent, "null", newclass);
}

