/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/
 
#include "kite_object.h"
#include "vm/kite_vm.h"

#define TYPECHECK_RIGHT_OPERAND(t) \
    kite_object_t *rhs = args->builtin_data.listvalue.car; \
    kite_object_t *exc; \
    if (rhs->type != t) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Parameter is not of the requested type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define METHOD_ADD(name, param, doc, param1, desc1) \
    kite_add_method(thread, newclass, # name, \
        kite_new_method_compiled_with_docs(thread, Compiler_ ## name, doc, param, param1, desc1));
        
KITE_CLASS_METHOD(Compiler_compileString) {
    int curstack = thd->running_stack->length - 1;
    kite_object_t *ret;
    TYPECHECK_RIGHT_OPERAND(OBJ_STRING);
    (void)this;

    ret = kite_vm_compile_from_string(thd, rhs->builtin_data.stringvalue.string, 1);
    /*kite_vm_call_object(thd, ret, ret, kite_new_null(thd));*/
    
    if ((thd->running_stack->length - 1) != curstack) {
        (void)kite_vm_pop(thd);
        while((thd->running_stack->length - 1) != curstack) {
            (void)kite_vm_pop(thd);
        }
    }

    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(Compiler_compileFile) {
    int curstack = thd->running_stack->length - 1;
    kite_object_t *ret;
    TYPECHECK_RIGHT_OPERAND(OBJ_STRING);
    (void)this;

    ret = kite_vm_compile_from_file(thd, rhs->builtin_data.stringvalue.string);
    /*kite_vm_call_object(thd, ret, ret, kite_new_null(thd));*/
    
    if ((thd->running_stack->length - 1) != curstack) {
        (void)kite_vm_pop(thd);
        while((thd->running_stack->length - 1) != curstack) {
            (void)kite_vm_pop(thd);
        }
    }
    
    kite_vm_return(thd, ret);
}

KITE_MODULE_INITIALIZER(compiler) {
    /* Create new class object. inherits is typically NULL for packages. */
    kite_object_t *newclass = kite_new_class(thread, NULL, "compiler");
    kite_set_docstring(newclass, "User-level access to the Kite compiler.");
    
    /* Add methods */
    METHOD_ADD(compileString, 1, "Compiles the Kite code contained in a string.",
               "str", "The string to compile.");
    METHOD_ADD(compileFile, 1, "Compiles the Kite code contained in a file.",
               "file", "The name of the file to load.");
    
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "compiler", 1, NULL);
    kite_set_property(thread, parent, "compiler", newclass);
}

