/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "kite_object.h"

/*****************************************************************************
 * Count the number of items in a list.
 ****************************************************************************/
int kite_list_count(kite_thread_t *thd, kite_object_t *obj)
{
    int ret = 0;
    (void)thd; /* to suppress warnings */
    assert(obj->type == OBJ_LIST);
    do {
        if (obj->builtin_data.listvalue.car) ret++;
        obj = obj->builtin_data.listvalue.cdr;
    } while(obj != NULL);
    return ret;
}

/*****************************************************************************
 * Append element to the end of the list.
 ****************************************************************************/
void 
kite_append_list(kite_thread_t *thd, kite_object_t *list, kite_object_t *obj)
{
    while(list->builtin_data.listvalue.car &&
          list->builtin_data.listvalue.cdr) {
loop:
        list = list->builtin_data.listvalue.cdr;
    }

    if (!list->builtin_data.listvalue.car) {
        list->builtin_data.listvalue.car = kite_reference_object(obj);
    } else {
        list->builtin_data.listvalue.cdr = kite_new_list(thd);
        goto loop;
    }
}
