/*****************************************************************************
 * Copyright (c) 2008, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#ifndef __MICROREGEX_INTERNAL
#define __MICROREGEX_INTERNAL

#include "microregex.h"

/******************************************************************************
 * NFA state object
 *****************************************************************************/
struct microregex_nfa_t
{
    enum {
        PUSH_SUBMATCH,
        POP_SUBMATCH,
        CHARACTER_CLASS,
        ALTERATION,
        ASSERT_BEGINNING,
        ASSERT_END,
        COUNTED_BEGIN,
        COUNTED_LOOP,
        COUNTED_END,
        BACKTRACK_CHECK,
        PASS,
        WORD_BOUNDARY,
        REGEX_END,
    } state_type;
    
    char *character_class;
    int invert, loop_from, loop_to;
    int ref;

    struct microregex_nfa_t *next1, *next2;
    microregex_nfa_t prevL, nextL;
};

/******************************************************************************
 * Main regular expression object.
 *****************************************************************************/
struct microregex_t
{
    const char *regex;
    int paren_count;
    int max_paren_count;
    char *err_string;
    int case_insensitive;
    int parser_flags;
    microregex_nfa_t nfa;
    microregex_nfa_t nlist;
};

/******************************************************************************
 * Regular expression state object
 *****************************************************************************/
struct microregex_state_t
{
    microregex_nfa_t state;
    int insensitive;
    int *se_stack;
    int se_size;
    char **ses;
    int num_ses;
    int match_begin; /* to facilitate split/replace */
    int match_end;
    int loop_count;
    struct microregex_state_t *prev, *next;
    struct microregex_state_t *prevL, *nextL; /* allocation list, used to free all re's */
};

#endif /* __MICROREGEX_INTERNAL */
