/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <assert.h>
#include "objs/kite_object.h"
#include "kite_vm.h"

/*****************************************************************************
 * Add item to list.
 ****************************************************************************/
static void kite_list_add_item(kite_list_t *location, kite_list_t *item) {
    kite_list_t *oldnext = location->next;

    item->next = oldnext;
    item->prev = location;
    if (item->next) item->next->prev = item;
    item->prev->next = item;
}

/*****************************************************************************
 * Remove item from the list.
 ****************************************************************************/
void 
kite_list_remove(kite_list_t **begin, kite_list_t **end, kite_list_t *item) {
    if (*begin == *end && *begin == item) {
        *begin = *end = NULL;
    } else if (*begin == item) {
        *begin = (*begin)->next;
        (*begin)->prev = NULL;
    } else if (*end == item) {
        *end = (*end)->prev;
        (*end)->next = NULL;
    } else {
        if (item->prev == item->next && item->next == NULL) return;
        item->prev->next = item->next;
        item->next->prev = item->prev;
    }
}

/*****************************************************************************
 * Add item to end of list.
 ****************************************************************************/
void 
kite_list_add_end(kite_list_t **begin, kite_list_t **end, kite_list_t *obj) {
    assert(begin != NULL && end != NULL && obj != NULL);

    if (*begin == *end && *end == NULL) {
        *begin = *end = obj;
    } else {
        kite_list_add_item(*end, obj);
        *end = obj;
    }
}

/*****************************************************************************
 * Add item to beginning of list.
 ****************************************************************************/
void 
kite_list_add_begin(kite_list_t **begin, kite_list_t **end, kite_list_t *obj)
{
    assert(begin != NULL && end != NULL && obj != NULL);

    if (*begin == *end && *begin == NULL) {
        *begin = *end = obj;
    } else {
        obj->next = *begin;
        *begin = obj;
        obj->next->prev = *begin;
    }
}

