%{
/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#define KITE_COMPILER
#include "objs/kite_object.h"
#include "kite_vm.h"

#define YYPARSE_PARAM parm
#define YYLEX_PARAM parm

/*
 * Dirty, dirty hack to make yyerror() work with Kite's exception system.
 * Credit: http://www.everything2.com/index.pl?node_id=1509675
 */
#define yyerror( msg )  dirty_yyerror( parm, msg )
void dirty_yyerror(void *, char *);
void checkReserved(void *, char *);

/* 
 * For some weird reason, we had to split the macro out this way
 * in order for the file/line numbering in exceptions to work properly.
 */
extern void *yyget_extra(void *);
extern int yylex(void *, void *, void *);

#define COMPILE_INSTRUCTION(i, j) \
    { \
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm); \
        kite_thread_t *thd = compiler->thd; \
        kite_object_t *cur = (kite_object_t*)kite_top_stack(compiler->obj_created); \
        COMPILE_INSTRUCTION2(i, j, parm, &cur->builtin_data.funcvalue.funcptr); \
        (void)thd; \
        (void)cur; \
    }
    
void COMPILE_INSTRUCTION2(kite_opcode_t *inst, int lno, void *p, void **dest) 
    { 
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(p); 
        kite_opcode_t *fptr = *dest;
        inst->file = strdup(compiler->file); 
        inst->line = lno; 
        kite_add_to_instruction_list(&fptr, inst);
        *dest = fptr;
    }

static void kite_optimize_increment(kite_opcode_t *o1, kite_opcode_t *o2, void *p)
{
    kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(p);
    kite_thread_t *thd = compiler->thd;
    kite_opcode_push *p1 = (kite_opcode_push*)o1;
    kite_opcode_push *p2 = (kite_opcode_push*)o2;
    if (o1->opcode == PUSH && o2->opcode == PUSH) {
        if (p1->obj->type == OBJ_IDENT && p2->obj->type == OBJ_IDENT) {
            if (strcmp(p1->obj->builtin_data.stringvalue.string,
                       p2->obj->builtin_data.stringvalue.string) == 0) {
                if (o1->next && o2->next) {
                    if (o1->next->opcode == DEREF_1 && o2->next->opcode == DEREF_1) {
                        if (o1->next->next == o2) {
                            kite_opcode_t *dupe = kite_compile_dupe_top(TRUE);
                            dupe->file = strdup(o2->file);
                            dupe->line = o2->line;
                            o1->next->next = dupe;
                            dupe->next = o2->next->next;
                            o2->next->next = NULL;
                            kite_free_instruction_list(thd, o2);
                        }
                    }
                }
            }
        }
    }
}
%}

%token IMPORT
%token METHOD
%token SYMBOL_NAME
%token FROM
%token CLASS
%token CONSTRUCT
%token DESTRUCT
%token PROPERTY
%token GLOBAL
%token PLUS
%token MINUS
%token MULTIPLY
%token DIVIDE
%token UNARY_PLUS
%token UNARY_MINUS
%token MAP
%token REDUCE
%token ARRAY_DEREF
%token EQUALS
%token NOT_EQUALS
%token LESS_THAN
%token GREATER_THAN
%token LESS_OR_EQUALS
%token GREATER_OR_EQUALS
%token AND
%token OR
%token NOT
%token STRING_VALUE
%token FLOAT_VALUE
%token INT_VALUE
%token MOD
%token OPERATOR
%token XOR
%token LEFT_SHIFT
%token RIGHT_SHIFT
%token EVAL
%token DECIDE
%token BOOL_VALUE
%token UNTIL
%token WHILE
%token RUN
%token CATCH
%token IS
%token ISOF
%token MAKE
%token METHOD_CALL
%token NULL_VALUE
%token DOCSTRING
%token BREAK
%token CONTINUE
%token RETURN
%token VERSION_KEYWORD
%token REGEX_VALUE
%token REGEX_VALUE_CI
%token ARRAY_SET

/* Needed since we can't quite lay out the associativity using the rules yet. */
%left MAP
%left REDUCE

%error-verbose
%defines
%pure-parser
%locations
%union {
    double doubleValue;
    long intValue;
    char *stringValue;
    struct kite_object_t *internalValue;
    kite_opcode_t *opValue;
}

%%

input:    /* empty */
    | DOCSTRING {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_object_t *obj = kite_top_stack(compiler->obj_created);
        kite_set_docstring(obj, $<stringValue>1);
        free($<stringValue>1);
    } input 
    | VERSION_KEYWORD STRING_VALUE ';' {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        char *tmp;
        
        if (compiler->verMajor)
        {
            yyerror("Cannot specify required version more than once.");
            YYERROR;
        }
        
        compiler->verMajor = atoi($<stringValue>2);
        tmp = strstr($<stringValue>2, ".");
        if (!tmp)
        {
            yyerror("Parse error while reading version string");
            YYERROR;
        }
        compiler->verPatch = atoi(tmp);
        tmp = strstr(tmp, ".");
        if (!tmp)
        {
            compiler->verPatch = 0;
        }
        
        if (compiler->verMajor > VERSION_MAJOR ||
           (compiler->verMajor == VERSION_MAJOR && compiler->verMinor > VERSION_MINOR) ||
           (compiler->verMajor == VERSION_MAJOR && compiler->verMinor == VERSION_MINOR && compiler->verPatch > VERSION_PATCH))
        {
            char buf[1024];
#ifndef WIN32
            snprintf(buf, 1024, "Module can only run in Kite version %d.%d.%d and above.",
                    compiler->verMajor, compiler->verMinor, compiler->verPatch);
#else
            sprintf_s(buf, 1024, "Module can only run in Kite version %d.%d.%d and above.",
                    compiler->verMajor, compiler->verMinor, compiler->verPatch);
#endif /* WIN32 */
            yyerror(buf);
            YYERROR;    
        }
    } input
    | statement ';' input
    ;

statement:
    | data_manipulation_statement
    | loop_statement
    | from_statement
    | import_statement
    | break_statement
    | continue_statement
    | return_statement
    | class_definition
    | property_definition
    | decide_statement
    | run_catch_statement
    | constructor_definition
    | destructor_definition
    | operator_definition
    ;
    
import_statement:    IMPORT STRING_VALUE     {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *obj = kite_top_stack(compiler->obj_created);

#ifndef HAVE_GC_H
        kite_funccall_info_t *fcall = (kite_funccall_info_t*)malloc(sizeof(kite_funccall_info_t));
#else
        kite_funccall_info_t *fcall = (kite_funccall_info_t*)GC_malloc(sizeof(kite_funccall_info_t));
#endif /* HAVE_GC_H */        
        fcall->file = compiler->file;
        fcall->line = @2.first_line;
        fcall->locals = NULL;
        fcall->last_deref = NULL;
        fcall->last_deref_obj = NULL;
        fcall->sys_or_user = 0;
        fcall->this = obj;
        kite_push_stack(&thd->func_stack, FALSE, fcall);
        
        kite_dereference_and_load(thd, $<stringValue>2);
        free($<stringValue>2);

        if (thd->exception) return 0;
        
        kite_pop_stack(&thd->func_stack);
#ifndef HAVE_GC_H
        free(fcall);
#endif /* HAVE_GC_H */
    }
    ;

break_statement:    BREAK
    {
        kite_opcode_t *opc = kite_compile_jump_uncond(NULL);
        opc->opcode = TEMP_BREAK;
        COMPILE_INSTRUCTION(opc, @1.first_line);
    }
    ;

continue_statement: CONTINUE
    {
        kite_opcode_t *opc = kite_compile_jump_uncond(NULL);
        opc->opcode = TEMP_CONTINUE;
        COMPILE_INSTRUCTION(opc, @1.first_line);
    }
    ;

return_statement: RETURN
    {
        kite_opcode_t *opc = kite_compile_return();
        COMPILE_INSTRUCTION(opc, @1.first_line);
    }
    ;
        
optional_docstring:
    {
        $<stringValue>$ = strdup("");
    }
    | DOCSTRING {
        $<stringValue>$ = $<stringValue>1;
    }
    ;
        
method_declaration: METHOD SYMBOL_NAME optional_param_name_list optional_docstring {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *method = 
            kite_new_method_bytecode(thd, NULL, 
                kite_list_count(thd, $<internalValue>3));
        kite_push_stack(&compiler->obj_created, FALSE,
                        method);
        kite_set_arginfo(method, $<internalValue>3);
        COMPILE_INSTRUCTION(kite_compile_funcargs($<internalValue>3), @1.first_line);
    } '[' input ']' {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *obj = kite_pop_stack(&compiler->obj_created);
        kite_object_t *parent = (kite_object_t*)kite_top_stack(compiler->obj_created);
        obj->parent = kite_reference_object(parent);
        kite_add_method(thd, parent, $<stringValue>2, obj);
        kite_set_docstring(obj, $<stringValue>4);
        free($<stringValue>2);
        free($<stringValue>4);
    }
    ;

anon_method_declaration: METHOD optional_param_name_list     {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *method = 
            kite_new_method_bytecode(thd, NULL, 
                kite_list_count(thd, $<internalValue>2));
        kite_push_stack(&compiler->obj_created, FALSE,
                        method);
        kite_set_arginfo(method, $<internalValue>2);
        COMPILE_INSTRUCTION(kite_compile_funcargs($<internalValue>2), @1.first_line);
    } '[' input ']' {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_object_t *obj = kite_pop_stack(&compiler->obj_created);
        COMPILE_INSTRUCTION(kite_compile_push(obj), @4.first_line);
    }
    ;
        
optional_global:    {
        $<intValue>$ = 0;
    }
    | GLOBAL    {
        $<intValue>$ = 1;
    }
    ;
    
optional_param_name_list:    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        $<internalValue>$ = kite_new_list(thd);
    }
    | param_name_list    {
        $<internalValue>$ = $<internalValue>1;
    }
    ;
    
param_name_list:    '(' ')'    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        $<internalValue>$ = kite_new_list(thd);
    }
    | '(' param_elements ')'    {
        $<internalValue>$ = $<internalValue>2;
    }
    ;
    
param_elements: SYMBOL_NAME optional_docstring   {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        checkReserved(parm, $<stringValue>1);
        $<internalValue>$ = kite_new_list(thd);
        $<internalValue>$->builtin_data.listvalue.car = 
            kite_new_ident(thd, $<stringValue>1);
        kite_set_docstring($<internalValue>$->builtin_data.listvalue.car, $<stringValue>2);
        free($<stringValue>1);
        free($<stringValue>2);
    }
    | SYMBOL_NAME optional_docstring ',' param_elements {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        checkReserved(parm, $<stringValue>1);
        $<internalValue>$ = kite_new_list(thd);
        $<internalValue>$->builtin_data.listvalue.car = 
            kite_new_ident(thd, $<stringValue>1);
        $<internalValue>$->builtin_data.listvalue.cdr = $<internalValue>4;
        kite_set_docstring($<internalValue>$->builtin_data.listvalue.car, $<stringValue>2);
        free($<stringValue>1);
        free($<stringValue>2);
    }
    ;
    
class_definition:    CLASS SYMBOL_NAME {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_push_stack(&compiler->obj_created, FALSE,
                         kite_new_class(thd, NULL, $<stringValue>2));
    } optional_from optional_docstring '[' class_body ']'    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *obj = kite_pop_stack(&compiler->obj_created);
        kite_funccall_info_t *fcall;
        kite_object_t *parent;

        obj->builtin_data.funcvalue.functype = FUNC_BYTECODE;
        parent = (kite_object_t*)kite_top_stack(compiler->obj_created);
        kite_add_property(thd, parent, $<stringValue>2, 1, NULL);
        kite_set_property(thd, parent, $<stringValue>2, obj);
        kite_set_docstring(obj, $<stringValue>5);
        free($<stringValue>2);
        free($<stringValue>5);
        
#ifndef HAVE_GC_H
        fcall = (kite_funccall_info_t*)malloc(sizeof(kite_funccall_info_t));
#else
        fcall = (kite_funccall_info_t*)GC_malloc(sizeof(kite_funccall_info_t));
#endif /* HAVE_GC_H */
        
        fcall->file = compiler->file;
        fcall->line = @1.first_line;
        fcall->locals = NULL;
        fcall->sys_or_user = 0;
        fcall->this = obj;
        fcall->last_deref = NULL;
        fcall->last_deref_obj = NULL;
        kite_push_stack(&thd->func_stack, FALSE, fcall);
        
        kite_vm_call_object(thd, obj, obj, kite_new_list(thd));
        
        if (thd->exception) return 0;
        
        kite_pop_stack(&thd->func_stack);
        kite_destruct_symtab(thd, &fcall->locals, TRUE);
#ifndef HAVE_GC_H
        free(fcall);
#endif /* HAVE_GC_H */
    }
    ;
    
optional_from:        {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *obj = (kite_object_t*)kite_top_stack(compiler->obj_created);
        obj->object_data.inherit_from = kite_dereference_and_load(thd, "System.object");
    }
    | from_statement
    ;

from_statement:
    FROM variable_immediate    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_object_t *obj = (kite_object_t*)kite_top_stack(compiler->obj_created);
        obj->object_data.inherit_from = $<internalValue>2;
        assert(obj->object_data.inherit_from->type == OBJ_CLASS ||
               obj->object_data.inherit_from->type == OBJ_INSTANCE);
    }
    ;
    
variable_immediate: SYMBOL_NAME {
#ifndef HAVE_GC_H
        kite_funccall_info_t *fcall = (kite_funccall_info_t*)malloc(sizeof(kite_funccall_info_t));
#else
        kite_funccall_info_t *fcall = (kite_funccall_info_t*)GC_malloc(sizeof(kite_funccall_info_t));
#endif /* HAVE_GC_H */
        void *opc = NULL;
        int idx;

        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        COMPILE_INSTRUCTION2(kite_compile_push(kite_new_ident(thd, $<stringValue>1)), 
                             @1.first_line, parm, &opc);
        COMPILE_INSTRUCTION2(kite_compile_deref_1(TRUE), @1.first_line, parm, &opc);
        
        fcall->file = compiler->file;
        fcall->line = @1.first_line;
        fcall->locals = NULL;
        fcall->last_deref = NULL;
        fcall->last_deref_obj = NULL;
        fcall->sys_or_user = 0;
        idx = compiler->obj_created->length;
        fcall->this =
            (kite_object_t*)compiler->obj_created->stack
                [idx >= 2 ? idx - 2 : idx - 1].obj;
        kite_push_stack(&thd->func_stack, FALSE, fcall);
        
        kite_vm_execute_user_method(thd, opc);
        
        if (thd->exception) return 0;
        
        kite_pop_stack(&thd->func_stack);
#ifndef HAVE_GC_H
        free(fcall);
#endif /* HAVE_GC_H */
        
        $<internalValue>$ = kite_vm_pop(thd);
        free($<stringValue>1);
        kite_free_instruction_list(thd, opc);
    }
    | variable_immediate '.' SYMBOL_NAME {
        void *opc = NULL;
        int idx;
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
#ifndef HAVE_GC_H
        kite_funccall_info_t *fcall = (kite_funccall_info_t*)malloc(sizeof(kite_funccall_info_t));
#else
        kite_funccall_info_t *fcall = (kite_funccall_info_t*)GC_malloc(sizeof(kite_funccall_info_t));
#endif /* HAVE_GC_H */
        COMPILE_INSTRUCTION2(kite_compile_push($<internalValue>1), 
                             @2.first_line, parm, &opc);
        COMPILE_INSTRUCTION2(kite_compile_push(kite_new_ident(thd, $<stringValue>3)), 
                             @2.first_line, parm, &opc);
        COMPILE_INSTRUCTION2(kite_compile_deref_2(), @2.first_line, parm, &opc);

        fcall->file = compiler->file;
        fcall->line = @1.first_line;
        fcall->locals = NULL;
        fcall->last_deref = NULL;
        fcall->last_deref_obj = NULL;
        fcall->sys_or_user = 0;
        idx = compiler->obj_created->length;
        fcall->this =
            (kite_object_t*)compiler->obj_created->stack
                [idx >= 2 ? idx - 2 : idx - 1].obj;
        kite_push_stack(&thd->func_stack, FALSE, fcall);
        
        kite_vm_execute_user_method(thd, opc);
        
        if (thd->exception) return 0;
        
        kite_pop_stack(&thd->func_stack);
#ifndef HAVE_GC_H
        free(fcall);
#endif /* HAVE_GC_H */

        $<internalValue>$ = kite_vm_pop(thd);
        free($<stringValue>3);
        kite_free_instruction_list(thd, opc);
    }
    ;
    
class_body:    
    | class_elements
    ;

class_elements:    class_element
    | class_elements ',' class_element
    ;
    
class_element:    constructor_definition
    | destructor_definition
    | property_definition
    | operator_definition
    | class_definition
    | data_manipulation_statement
    ;

constructor_definition:    CONSTRUCT optional_param_name_list {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_push_stack(&compiler->obj_created, FALSE,
                         kite_new_method_bytecode(thd, NULL, 
                                        kite_list_count(thd, $<internalValue>2)));
        COMPILE_INSTRUCTION(kite_compile_funcargs($<internalValue>2), @1.first_line);
    } optional_docstring '[' input ']' {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *obj = kite_pop_stack(&compiler->obj_created);
        kite_object_t *parent = (kite_object_t*)kite_top_stack(compiler->obj_created);
        obj->parent = kite_reference_object(parent);
        kite_add_method(thd, parent, "__construct__", obj);
        kite_set_docstring(obj, $<stringValue>4);
        kite_set_arginfo(obj, $<internalValue>2);
        free($<stringValue>4);
    }
    ;
    
destructor_definition: DESTRUCT {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_push_stack(&compiler->obj_created, FALSE,
                         kite_new_method_bytecode(thd, NULL, 0));
    } optional_docstring '[' input ']' {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *obj = kite_pop_stack(&compiler->obj_created);
        kite_object_t *parent = (kite_object_t*)kite_top_stack(compiler->obj_created);
        obj->parent = kite_reference_object(parent);
        kite_add_method(thd, parent, "__destruct__", obj);
        kite_set_docstring(obj, $<stringValue>3);
        free($<stringValue>3);
    }
    ;
    
property_definition:    optional_global PROPERTY SYMBOL_NAME optional_docstring   {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *parent = (kite_object_t*)kite_top_stack(compiler->obj_created);
        kite_symtab_t *symt;
        
        if (parent->type == OBJ_METHOD) {
            COMPILE_INSTRUCTION(
                kite_compile_push(kite_new_ident(thd, $<stringValue>3)), @2.first_line);
            COMPILE_INSTRUCTION(kite_compile_defprop(), @2.first_line);
        } else {
            kite_add_property(thd, parent, $<stringValue>3, $<intValue>1, NULL);
            KITE_FIND_PROPERTY(symt, parent, $<stringValue>3, $<intValue>1);
            symt->docstring = strdup($<stringValue>4);
        }
        free($<stringValue>3);
        free($<stringValue>4);
    }
    | optional_global PROPERTY SYMBOL_NAME optional_docstring '=' {
            kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
            kite_thread_t *thd = compiler->thd;
            kite_object_t *parent = (kite_object_t*)kite_top_stack(compiler->obj_created);
            kite_symtab_t *symt;

            COMPILE_INSTRUCTION(
                kite_compile_push(kite_new_ident(thd, $<stringValue>3)), @2.first_line);
            COMPILE_INSTRUCTION(kite_compile_deref_1(FALSE), @2.first_line);
            if (parent->type == OBJ_METHOD) {
                COMPILE_INSTRUCTION(
                    kite_compile_push(kite_new_ident(thd, $<stringValue>3)), @2.first_line);
                COMPILE_INSTRUCTION(kite_compile_defprop(), @2.first_line);
            } else {
                kite_add_property(thd, parent, $<stringValue>3, $<intValue>1, NULL);
                KITE_FIND_PROPERTY(symt, parent, $<stringValue>3, $<intValue>1);
                symt->docstring = strdup($<stringValue>4);
            }
            free($<stringValue>3);
            free($<stringValue>4);
        } data_manipulation_statement {
            COMPILE_INSTRUCTION(kite_compile_arithop(OP_ASSIGNMENT), @5.first_line);
        }
    ;

operator_definition:    OPERATOR operator_type param_name_list {
    kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
    kite_thread_t *thd = compiler->thd;
    kite_push_stack(&compiler->obj_created, FALSE,
                     kite_new_method_bytecode(thd, NULL, 
                                    kite_list_count(thd, $<internalValue>3)));
    COMPILE_INSTRUCTION(kite_compile_funcargs($<internalValue>3), @1.first_line);
} optional_docstring '[' input ']' {
    kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
    kite_thread_t *thd = compiler->thd;
    kite_object_t *obj = kite_pop_stack(&compiler->obj_created);
    kite_object_t *parent = (kite_object_t*)kite_top_stack(compiler->obj_created);
    kite_add_operator(thd, parent, $<intValue>2, obj);
    kite_set_docstring(obj, $<stringValue>5);
    free($<stringValue>5);
}
    ;
    
operator_type:    PLUS    {
        $<intValue>$ = OP_ADD;
    }
    | MINUS    {
        $<intValue>$ = OP_SUBTRACT;
    }
    | MULTIPLY     {
        $<intValue>$ = OP_MULTIPLY;
    }
    | DIVIDE    {
        $<intValue>$ = OP_DIVIDE;
    }
    | MOD    {
        $<intValue>$ = OP_MODULUS;
    }
    | UNARY_PLUS    {
        $<intValue>$ = OP_UNARY_PLUS;
    }
    | UNARY_MINUS    {
        $<intValue>$ = OP_UNARY_MINUS;
    }
    | MAP    {
        $<intValue>$ = OP_MAP;
    }
    | REDUCE    {
        $<intValue>$ = OP_REDUCE;
    }
    | ARRAY_DEREF    {
        $<intValue>$ = OP_ARRAY_DEREF;
    }
    | ARRAY_SET {
        $<intValue>$ = OP_ARRAY_SET;
    }
    | EQUALS    {
        $<intValue>$ = OP_EQUALS;
    }
    | NOT_EQUALS    {
        $<intValue>$ = OP_NOT_EQUALS;
    }
    | LESS_THAN    {
        $<intValue>$ = OP_LESS_THAN;
    }
    | GREATER_THAN    {
        $<intValue>$ = OP_GREATER_THAN;
    }
    | LESS_OR_EQUALS    {
        $<intValue>$ = OP_LESS_OR_EQUALS;
    }
    | GREATER_OR_EQUALS    {
        $<intValue>$ = OP_GREATER_OR_EQUALS;
    }
    | AND    {
        $<intValue>$ = OP_AND;
    }
    | OR    {
        $<intValue>$ = OP_OR;
    }
    | NOT    {
        $<intValue>$ = OP_NOT;
    }
    | XOR    {
        $<intValue>$ = OP_XOR;
    }
    | LEFT_SHIFT    {
        $<intValue>$ = OP_LEFT_SHIFT;
    }
    | RIGHT_SHIFT    {
        $<intValue>$ = OP_RIGHT_SHIFT;
    }
    | METHOD_CALL    {
        $<intValue>$ = OP_METHOD_CALL;
    }
    | PROPERTY       {
        $<intValue>$ = OP_PROPERTY;
    }
    ;

data_manipulation_statement:    assignment_statement
    | method_declaration
    | eval_statement
    ;
    
assignment_statement:    deref_statement {
    kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
    kite_object_t *cur = (kite_object_t*)kite_top_stack(compiler->obj_created);
    kite_opcode_t *opc = cur->builtin_data.funcvalue.funcptr;
    
    while(opc->next) opc = opc->next;
    
    if (opc->opcode == ARITH_OP && ((kite_opcode_arithop*)opc)->operation == OP_ARRAY_DEREF)
        $<opValue>$ = opc;
    else
        $<opValue>$ = NULL;

} '=' data_manipulation_statement {
        kite_opcode_t *op1 = $<opValue>1;
        kite_opcode_t *op2;
        
        if ($<opValue>1 && $<opValue>1->next && $<opValue>1->next->next) 
        {
            op2 = $<opValue>1->next->next;
            kite_optimize_increment(op1, op2, parm);
        }
        
        if (!$<opValue>2)
        {
            COMPILE_INSTRUCTION(kite_compile_arithop(OP_ASSIGNMENT), @2.first_line);
        }
        else
        {
            COMPILE_INSTRUCTION(kite_compile_arithop(OP_ARRAY_SET), @2.first_line);
            $<opValue>2->opcode = NOP;
        }
    }
    | or_statement
    ;
    
or_statement:        xor_statement {
        COMPILE_INSTRUCTION(kite_compile_dupe_top(FALSE), @1.first_line);
        $<opValue>$ = kite_compile_jump_true(NULL);
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    } OR or_statement {
        kite_opcode_t *opc = kite_compile_nop();
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);

        (void)compiler;
        opc->file = "";
        opc->line = @3.first_line;

        COMPILE_INSTRUCTION(kite_compile_arithop(OP_OR), @3.first_line);
        ((kite_opcode_jump_true*)$<opValue>2)->jumpto = opc;
        COMPILE_INSTRUCTION(opc, @3.first_line);
    }
    | xor_statement
    ;

xor_statement:    xor_statement XOR and_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_XOR), @2.first_line);
    }
    | and_statement
    ;
    
and_statement:    bitwise_statement {
        COMPILE_INSTRUCTION(kite_compile_dupe_top(FALSE), @1.first_line);
        $<opValue>$ = kite_compile_jump_false(NULL);
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    } AND and_statement {
        kite_opcode_t *opc = kite_compile_nop();
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);

        (void)compiler;
        opc->file = "";
        opc->line = @3.first_line;
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_AND), @3.first_line);
        ((kite_opcode_jump_true*)$<opValue>2)->jumpto = opc;
        COMPILE_INSTRUCTION(opc, @3.first_line);
    }
    | bitwise_statement
    ;

bitwise_statement:    bitwise_statement LEFT_SHIFT equality_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_LEFT_SHIFT), @2.first_line);
    }
    | bitwise_statement RIGHT_SHIFT equality_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_RIGHT_SHIFT), @2.first_line);
    }
    | equality_statement
    ;
    
equality_statement:    equality_statement EQUALS less_greater_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_EQUALS), @2.first_line);
    }
    | equality_statement NOT_EQUALS less_greater_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_NOT_EQUALS), @2.first_line);
    }
    | equality_statement IS less_greater_statement {
        COMPILE_INSTRUCTION(kite_compile_objis(FALSE), @2.first_line);
    }
    | equality_statement ISOF less_greater_statement {
        COMPILE_INSTRUCTION(kite_compile_objis(TRUE), @2.first_line);
    }
    | less_greater_statement
    ;
    
less_greater_statement:    less_greater_statement '<' add_sub_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_LESS_THAN), @2.first_line);
    }
    | less_greater_statement '>' add_sub_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_GREATER_THAN), @2.first_line);
    }
    | less_greater_statement LESS_OR_EQUALS add_sub_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_LESS_OR_EQUALS), @2.first_line);
    }
    | less_greater_statement GREATER_OR_EQUALS add_sub_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_GREATER_OR_EQUALS), @2.first_line);
    }
    | add_sub_statement
    ;
    
add_sub_statement:    add_sub_statement '+' multdiv_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_ADD), @2.first_line);
    }
    | add_sub_statement '-' multdiv_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_SUBTRACT), @2.first_line);
    }
    | multdiv_statement
    ;
    
multdiv_statement:    multdiv_statement '*' unary_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_MULTIPLY), @2.first_line);
    }
    | multdiv_statement '/' unary_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_DIVIDE), @2.first_line);
    }
    | multdiv_statement '%' unary_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_MODULUS), @2.first_line);
    }
    | unary_statement
    ;
    
unary_statement: '+' unary_statement {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_null(thd)), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_UNARY_PLUS), @1.first_line);
    }
    | '-' unary_statement {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_null(thd)), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_UNARY_MINUS), @1.first_line);
    }
    | NOT unary_statement {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_null(thd)), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_NOT), @1.first_line);
    }
    | map_reduce_statement
    ;
    
map_reduce_statement:    deref_statement MAP deref_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_MAP), @2.first_line);
    }
    | deref_statement REDUCE deref_statement {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_REDUCE), @2.first_line);
    }
    | SYMBOL_NAME { 
        char *str = $<stringValue>1;        
		COMPILE_INSTRUCTION(kite_compile_this(), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_ident(thd, str)), @1.first_line);
        free(str);
        COMPILE_INSTRUCTION(kite_compile_deref_1(FALSE), @1.first_line);
    } param_list {
        COMPILE_INSTRUCTION(kite_compile_call(), @1.first_line);
    }
    | deref_statement
    ;
    
deref_statement:    deref_statement '|' SYMBOL_NAME {
        char *str = $<stringValue>3;
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_ident(thd, str)), @3.first_line);
        free(str);
    } optional_param_list {
        if ($<opValue>5->line == 0) $<opValue>5->line = @3.first_line;
        COMPILE_INSTRUCTION(kite_compile_call(), @3.first_line);
        $<opValue>$ = $<opValue>1;
    }
    | deref_statement '[' data_manipulation_statement ']'    {
        COMPILE_INSTRUCTION(kite_compile_arithop(OP_ARRAY_DEREF), @3.first_line);
        $<opValue>$ = $<opValue>1;
    }
    | deref_statement '.' SYMBOL_NAME {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_ident(thd, $<stringValue>3)),
             @2.first_line);
        COMPILE_INSTRUCTION(kite_compile_deref_2(), @2.first_line);
        free($<stringValue>3);
        $<opValue>$ = $<opValue>1;
    }
    | SYMBOL_NAME {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_opcode_t *op = kite_compile_push(kite_new_ident(thd, $<stringValue>1));
        COMPILE_INSTRUCTION(op, @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_deref_1(TRUE), @1.first_line);
        free($<stringValue>1);
        $<opValue>$ = op;
    }
    | constant_statement {
        $<opValue>$ = $<opValue>1;
    }
    ;

optional_param_list: {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        $<opValue>$ = kite_compile_push(kite_new_list(thd));
        COMPILE_INSTRUCTION($<opValue>$, 0);
    }
    | param_list {
        $<opValue>$ = $<opValue>1;
    }
    ;

param_list:    '(' ')' {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        $<opValue>$ = kite_compile_push(kite_new_list(thd));
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    }
    | '(' params_list ')' {
        $<opValue>$ = $<opValue>2;
    }
    ;

params_list: data_manipulation_statement    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        $<opValue>$ = kite_compile_list_cons_1(thd);
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    }
    | data_manipulation_statement ',' params_list    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        $<opValue>$ = kite_compile_list_cons(thd);
        COMPILE_INSTRUCTION($<opValue>$, @2.first_line);
    }
    ;
    
var_name:    SYMBOL_NAME {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_opcode_t *op = kite_compile_push(kite_new_ident(thd, $<stringValue>1));
        COMPILE_INSTRUCTION(op, @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_deref_1(TRUE), @1.first_line);
        free($<stringValue>1);
        $<opValue>$ = op;
    }
    | var_name '.' SYMBOL_NAME {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_ident(thd, $<stringValue>3)),
             @2.first_line);
        COMPILE_INSTRUCTION(kite_compile_deref_2(), @2.first_line);
        free($<stringValue>3);
        $<opValue>$ = $<opValue>1;
    }
    ;
    
constant_statement:    STRING_VALUE     {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_opcode_t *opc = kite_compile_push(kite_new_string(thd, $<stringValue>1));
        COMPILE_INSTRUCTION(opc, @1.first_line);
        free($<stringValue>1);
        $<opValue>$ = opc;
    }
    | INT_VALUE {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_opcode_t *opc = kite_compile_push(kite_new_integer(thd, $<intValue>1));
        COMPILE_INSTRUCTION(opc, @1.first_line);
        $<opValue>$ = opc;
    }
    | FLOAT_VALUE {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_opcode_t *opc = kite_compile_push(kite_new_float(thd, $<doubleValue>1));
        COMPILE_INSTRUCTION(opc, @1.first_line);
        $<opValue>$ = opc;
    }
    | BOOL_VALUE {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_opcode_t *opc = kite_compile_push(kite_new_boolean(thd, $<intValue>1));
        COMPILE_INSTRUCTION(opc, @1.first_line);
        $<opValue>$ = opc;
    }
    | NULL_VALUE {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_opcode_t *opc = kite_compile_push(kite_new_null(thd));
        COMPILE_INSTRUCTION(opc, @1.first_line);
        $<opValue>$ = opc;
    }
    | INT_VALUE ':' INT_VALUE
    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *range_class = kite_dereference_and_load(thd, "System.math.range");
        kite_object_t *range_args = kite_new_list(thd);
        kite_object_t *range_object;

        kite_append_list(thd, range_args, kite_new_integer(thd, $<intValue>1));
        kite_append_list(thd, range_args, kite_new_integer(thd, $<intValue>3));
        range_object = 
            kite_new_instance_with_constructor(thd, range_class, range_args);
        COMPILE_INSTRUCTION(kite_compile_push(range_object), @2.first_line);
        kite_dereference_object(range_args);
        kite_dereference_object(range_class);
        kite_dereference_object(range_object);
        $<opValue>$ = NULL; 
    }
    | INT_VALUE ':' INT_VALUE ':' INT_VALUE
    {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *range_class = kite_dereference_and_load(thd, "System.math.range");
        kite_object_t *range_args = kite_new_list(thd);
        kite_object_t *range_object;

        kite_append_list(thd, range_args, kite_new_integer(thd, $<intValue>1));
        kite_append_list(thd, range_args, kite_new_integer(thd, $<intValue>3));
        kite_append_list(thd, range_args, kite_new_integer(thd, $<intValue>5));
        range_object = 
            kite_new_instance_with_constructor(thd, range_class, range_args);
        COMPILE_INSTRUCTION(kite_compile_push(range_object), @2.first_line);
        kite_dereference_object(range_args);
        kite_dereference_object(range_class);
        kite_dereference_object(range_object);
        $<opValue>$ = NULL;
    }
    | REGEX_VALUE {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *param_list = kite_new_list(thd);
        kite_object_t *regex_obj = kite_dereference_and_load(thd, "System.regex");
        
        kite_append_list(thd, param_list, kite_new_string(thd, $<stringValue>1));
        
        COMPILE_INSTRUCTION(kite_compile_push(regex_obj), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_push(param_list), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_make(), @1.first_line);
        
        free($<stringValue>1);
        $<opValue>$ = NULL;
    }
    | REGEX_VALUE_CI {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *param_list = kite_new_list(thd);
        kite_object_t *regex_obj = kite_dereference_and_load(thd, "System.regex");
        
        kite_append_list(thd, param_list, kite_new_string(thd, $<stringValue>1));
        kite_append_list(thd, param_list, kite_new_boolean(thd, TRUE));
        
        COMPILE_INSTRUCTION(kite_compile_push(regex_obj), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_push(param_list), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_make(), @1.first_line);
        
        free($<stringValue>1);
        $<opValue>$ = NULL;
    }
    | list_value { $<opValue>$ = NULL; }
    | construct_object_statement { $<opValue>$ = NULL; }
    | anon_method_declaration { $<opValue>$ = NULL; }
    | '(' data_manipulation_statement ')' { $<opValue>$ = NULL; }
    ;

construct_object_statement:    MAKE var_name param_list    {
        COMPILE_INSTRUCTION(kite_compile_make(), @1.first_line);
    }
    ;
    
list_value: '[' optional_list_elements ']'
    ;
    
optional_list_elements:    {
        COMPILE_INSTRUCTION(kite_compile_list_cons_0(thd), 0);
    }
    | list_elements
    ;
    
list_elements: list_element    {
        COMPILE_INSTRUCTION(kite_compile_list_cons_1(thd), @1.first_line);
    }
    | list_element ',' list_elements    {
        COMPILE_INSTRUCTION(kite_compile_list_cons(thd), @2.first_line);
    }
    ;

list_element:    or_statement
    ;
    
decide_statement:    DECIDE {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_push_stack(&compiler->decide_exits, FALSE, kite_compile_nop());
    } '[' decide_contents ']' {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_opcode_t *tmpopc = kite_pop_stack(&compiler->decide_exits);
        COMPILE_INSTRUCTION(tmpopc, @5.first_line);
    }
    ;
    
decide_contents:    decide_condition
    | decide_contents ',' decide_condition
    ;
    
decide_condition:    constant_statement {
        $<opValue>$ = kite_compile_jump_false(NULL);
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    } '[' input ']'    {
        kite_opcode_jump_false *c = (kite_opcode_jump_false*)$<opValue>2;
        c->jumpto = kite_compile_nop();
        
        COMPILE_INSTRUCTION(
            kite_compile_jump_uncond(
                kite_top_stack(compiler->decide_exits)
            ),
            @3.first_line
        );
        
        COMPILE_INSTRUCTION(c->jumpto, @3.first_line);
    }
    ;
    
eval_statement:    EVAL {
        kite_compiler_t *compiler = (kite_compiler_t*)yyget_extra(parm);
        kite_thread_t *thd = compiler->thd;
        kite_object_t *c = kite_dereference_and_load(thd, "System.vm.compiler");
        COMPILE_INSTRUCTION(kite_compile_push(c), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_ident(thd, "compileString")), 
            @1.first_line);
    } or_statement    {
        COMPILE_INSTRUCTION(kite_compile_list_cons_1(thd), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_call(), @1.first_line);
        COMPILE_INSTRUCTION(kite_compile_eval(), @1.first_line);
    }
    ;

loop_statement:    loop_type '[' input ']'    {
        kite_opcode_t *loop_bottom_cont = kite_compile_jump_uncond($<opValue>1);
        kite_opcode_t *loop_bottom_end = kite_compile_nop();
        kite_opcode_t *opc;

        COMPILE_INSTRUCTION(loop_bottom_cont, @4.first_line);
        
        /* Fix jump destination */
        opc = $<opValue>1;
        while(opc && opc != loop_bottom_cont)
        {
            if ((opc->opcode == JUMP_TRUE && ((kite_opcode_jump_true*)opc)->jumpto == NULL) ||
                (opc->opcode == JUMP_FALSE && ((kite_opcode_jump_false*)opc)->jumpto == NULL)) 
            {
                ((kite_opcode_jump_true*)opc)->jumpto = loop_bottom_end;
            }
            else if (opc->opcode == TEMP_BREAK)
            {
                opc->opcode = JUMP_UNCOND;
                ((kite_opcode_jump_true*)opc)->jumpto = loop_bottom_end;
            }
            else if (opc->opcode == TEMP_CONTINUE)
            {
                opc->opcode = JUMP_UNCOND;
                ((kite_opcode_jump_true*)opc)->jumpto = loop_bottom_cont;
            }
            opc = opc->next;
        }
        
        COMPILE_INSTRUCTION(loop_bottom_end, @4.first_line);
    }
    ;

loop_type:    WHILE {
        $<opValue>$ = kite_compile_nop();
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    } constant_statement    {
        COMPILE_INSTRUCTION(kite_compile_jump_false(NULL), @2.first_line);
        $<opValue>$ = $<opValue>2;
    }
    | UNTIL {
        $<opValue>$ = kite_compile_nop();
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    } constant_statement    {
        COMPILE_INSTRUCTION(kite_compile_jump_true(NULL), @2.first_line);
        $<opValue>$ = $<opValue>2;
    }
    ;

run_catch_statement: RUN {
        $<opValue>$ = kite_compile_exception_hndl(NULL);
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    } '[' input ']' {
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_boolean(thd, TRUE)), @1.first_line);
        $<opValue>$ = kite_compile_jump_true(NULL);
        COMPILE_INSTRUCTION($<opValue>$, @1.first_line);
    } CATCH {
        kite_opcode_t *opc = kite_compile_nop();
        ((kite_opcode_exception_hndl*)$<opValue>2)->handler = opc;
        COMPILE_INSTRUCTION(opc, @7.first_line);
    } '[' input ']' {
        kite_opcode_t *opc = kite_compile_pop_exception_hndl();
        kite_opcode_t *nop = kite_compile_nop();
        ((kite_opcode_jump_true*)$<opValue>6)->jumpto = opc;
        COMPILE_INSTRUCTION(kite_compile_push(kite_new_boolean(thd, TRUE)), @7.first_line);
        COMPILE_INSTRUCTION(kite_compile_jump_true(nop), @7.first_line);
        COMPILE_INSTRUCTION(opc, @7.first_line);
        COMPILE_INSTRUCTION(nop, @7.first_line);
    }
    ;
    
%%

void dirty_yyerror(void *yyparse_param, char *str) {
    kite_object_t *exc;
    kite_compiler_t *comp = (kite_compiler_t*)yyget_extra(yyparse_param);
    kite_thread_t *thd = comp->thd;
    
    kite_funccall_info_t *entry = calloc(sizeof(kite_funccall_info_t), 1);
    entry->this = NULL;
    entry->file = comp->file;
    entry->line = comp->currentLine;
    kite_push_stack(&thd->func_stack, FALSE, entry);

    exc = kite_new_exception(thd, "System.exceptions.SyntaxError", 
                             str);
    kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);    
}

void checkReserved(void *p, char *name) {
    if (!strcmp(name, "__exc") ||
        !strcmp(name, "this")) {
        dirty_yyerror(p, "Cannot use reserved names for parameters.");
    }
}
